/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>

#include "s68.h"
#include "device.h"
#include "conf.h"


static void epromLoad(char *fName, u_char *eprom, u_long base, u_long size);



/*
 * e p r o m R e s e t
 *
 * Reset the device
 */
int epromReset(Device *dev)
{
    if (!quiet)  uiPrintf(" EPROM");

    return 0;
}



/*
 * e p r o m R e a d
 *
 * Read one byte
 */
static u_long epromRead(Device *dev, u_long offset)
{
    u_char *eprom = (u_char *) dev->priv;

    return (u_long) eprom[offset];
}



/*
 * e p r o m W r i t e
 *
 * Write one byte
 */
static void epromWrite(Device *dev, u_long offset, u_char data)
{
#ifdef DEBUG
    u_char *eprom = (u_char *) dev->priv;

    eprom[offset] = data;
#endif
}



/*
 * i n i t E P R O M
 *
 * Initialize the device
 */
int initEPROM(Device *dev, u_long base, u_long size, void *arg)
{
    u_char  *eprom;
	u_long  i;

    if ((eprom = malloc(size)) == NULL) {
        uiErrPrintf("EPROM: Malloc of %ld bytes failed\n", size);
        return(errno);
    }

    for (i = 0; i < size; i++)  eprom[i] = 0xFF;
    if (arg != NULL)
        epromLoad((char *) arg, eprom, base, size);

    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) eprom;
    dev->devReset = epromReset;
    dev->devRead = epromRead;
    dev->devWrite = epromWrite;
    dev->devIAck  = NULL;
    dev->devCycle = NULL;

    return 0;
}



/*
 * e p r o m R e a d L i n e
 *
 * Read line from file.
 */
static char *epromReadLine(char *bp, int sz, FILE *f)
{
    int  c, i;

    for (i = 0; i < sz; i++)  {
        c = getc(f);
        if ((c == '\r') || (c == '\n')) {
            bp[i] = '\0';
            break;
        }
        bp[i] = (char) c;
    }

    return bp;
}



/*
 * e p r o m L o a d S R e c
 *
 * Load memory with contents of S-Record file
 */
static int epromLoadSRec(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    char   buffer[128];
    char   *bptr, type;
    int    line = 0;	
    u_long i;
    u_long address, nbyte, data;


    if ((fp = fopen(fName, "r")) == NULL)  {
        uiErrPrintf("EPROM: Can't open %s - %s\n", fName, strerror(errno));
        return(errno);
    }

    while (!feof(fp))  {
        line++;
        if ((bptr = epromReadLine(buffer, 128, fp)) == NULL)  {
            uiErrPrintf("EPROM: Unexpected end of file - %s\n", strerror(errno));
            return(-1);
        }

        /* Get Type */
        if (sscanf(bptr, "S%c", &type) != 1)  {
            /* Ignore lines that don't start with S */
            break;
        }
        bptr += 2;

        /* Get Length */
        if (sscanf(bptr, "%02lx", &nbyte) != 1)  {
            uiErrPrintf("EPROM: Line %d: Bad character in length\n", line);
            return(-1);
        }
        bptr += 2;

        /* Get address field */
        switch (type)  {
            case '0' :
                /* Ignore start record */
                continue;

            case '1' :
                if (sscanf(bptr,"%04lx", &address) != 1)  {
                    uiErrPrintf("EPROM: Line %d: Bad address\n", line);
                    return(-1);
                }
                bptr += 4;
                nbyte -= 2;
                break;

            case '2' :
                if (sscanf(bptr,"%06lx", &address) != 1)  {
                    uiErrPrintf("EPROM: Line %d: Bad address\n", line);
                    return(-1);
                }
                bptr += 6;
                nbyte -= 3;
                break;

            case '3' :
                if (sscanf(bptr,"%08lx", &address) != 1)  {
                    uiErrPrintf("EPROM: Line %d: Bad address\n", line);
                    return(-1);
                }
                bptr += 8;
                nbyte -= 4;
                break;

            case '7' :
            case '8' :
            case '9' :
                /* Ignore all termination records */
                continue;

            default :
                uiErrPrintf("EPROM: Line %d: Bad type '%c'\n", line, type);
                return(-1);
        }

        /* Get all data bytes without checksum */
        for (i = 0; i < nbyte-1; i++)  {
            if (sscanf(bptr, "%02lx", &data) != 1)  {
                uiErrPrintf("EPROM: Line %d: Bad character in data\n", line);
                return(-1);
            }
            bptr += 2;
            if ((address >= base) || (address < (base + size)))  {
                eprom[address - base] = (u_char) data;
                address++;
            }
            else  {
                uiErrPrintf("EPROM: Address %08lx out of range", address);
                return(-1);
            }
        }

        /* Ignore checksum */
        if (sscanf(bptr, "%02lx", &data) != 1)  {
            uiErrPrintf("EPROM: Line %d: Bad character in checksum\n", line);
            return(-1);
        }
    }

    fclose(fp);

	return 0;
}



/*
 * e p r o m L o a d B i n
 *
 * Load memory with contents of Binary file
 */
static int epromLoadBin(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    char   buffer[128];
    u_long i;
    u_long address = 0;
    u_long nbyte;


    if ((fp = fopen(fName, "rb")) == NULL)  {
        uiErrPrintf("EPROM: Can't open %s - %s\n", fName, strerror(errno));
        return(errno);
    }

    while (!feof(fp))  {
        nbyte = fread(buffer, sizeof(u_char), 128, fp);
        for (i = 0; i < nbyte; i++)  {
            if (address >= size)  {
                uiErrPrintf("EPROM: Reached end of EPROM\n");
                return(-1);
            }
            eprom[address++] = (u_char) buffer[i];
        }
    }

    fclose(fp);

	return 0;
}



/*
 * e p r o m L o a d
 *
 * Load memory with contents of Binary file
 */
static void epromLoad(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    int    c;
	int    ret;


    if ((fp = fopen(fName, "r")) == NULL)  {
        uiErrPrintf("EPROM: Can't open %s\n", fName);
        return;
    }

    c = getc(fp);
    fclose(fp);

    if (c == (int) 'S')
        ret = epromLoadSRec(fName, eprom, base, size);
    else
        ret = epromLoadBin(fName, eprom, base, size);
}
