/*
 * (c) Copyright 2002-2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "s68.h"
#include "device.h"
#include "decoder.h"
#include "conf.h"
#include "ui.h"

#define REGSIZE   16


#define MSR           0x01       /* Main Status Register (read only) */
#define DATA          0x03       /* Data Register */
#define TERMCNTL      0x05
#define MOTCNTL       0x09
#define INTEN         0x0d

#define MSR_DRV1BUSY  0x02       /* Drive 1 Busy */
#define MSR_DRV2BUSY  0x04       /* Drive 2 Busy */
#define MSR_DRV3BUSY  0x08       /* Drive 3 Busy */
#define MSR_CIP       0x10       /* Command in Progress */
#define MSR_EXC       0x20       /* Non-DMA Execution */
#define MSR_DIO       0x40       /* Data I/O (Direction) */
#define MSR_RQM       0x80       /* Request for Master */


/* Controller Phase Definitions */
#define PH_IDLE       0
#define PH_COMMAND    1
#define PH_EXECUTION  2
#define PH_RESULT     3

/* Command Definitions */
#define CMD_SPECIFY       0x03
#define CMD_SENSE_DRVSTAT 0x04
#define CMD_WRITE_DATA    0x05
#define CMD_READ_DATA     0x06
#define CMD_RECAL         0x07
#define CMD_SENSE_INT     0x08
#define CMD_FORMAT_TRK    0x0D
#define CMD_SEEK          0x0F
#define CMD_VERSION       0x10


typedef struct  {
    u_char  msr;
    u_char  dsr;
    u_char  data;

    u_char  sectorBuffer[1024];
    int     sectorIdx;
    u_long  diskSize;

    int  numOfHeads;
    int  numOfSecpTrk;
    int  numOfCylinder;

    int    phase;                /* Controller command phase */
    u_char cmd[12];
    int    cidx, ccnt;
    u_char stat[8];
    int    sidx, scnt;

    int    curTrack;             /* Current track number starting at 0 */
    int    curHead;              /* Current head number 0 or 1*/
    int    curSector;            /* Current secotr number starting at 1 */
    int    secTrack;             /* Sectors per Track */
    int    secSize;              /* Sector size */
    int    secCount;             /* Sector counter used for format track */

    u_char idBuffer[8];
    int    idIdx;
    u_char dataPat;

    FILE   *imgFile;
} NEC765;





/*
 * t r a n s f e r S e c t o r
 *
 * Transfer sector buffer to/from disk
 */
static int transferSector(NEC765 *nec765, u_char mode)
{
    u_long  lba;
    u_long  fileOffset;
    off_t   ret;
    size_t  cnt;

    lba = (nec765->curTrack * nec765->numOfHeads * nec765->secTrack)
        + (nec765->curHead * nec765->secTrack)
        + (nec765->curSector - 1);
    fileOffset = lba * nec765->secSize;

    ret = fseek(nec765->imgFile, fileOffset, SEEK_SET);
    if (ret == (off_t) -1)  {
        /* Seek error */
        uiErrPrintf("NEC765: Image file seek error - %s\n", strerror(errno));
        return(-1);
    }

    switch (mode)  {
      case CMD_READ_DATA :
          cnt = fread(nec765->sectorBuffer, 1, nec765->secSize, nec765->imgFile);
          if (cnt != (size_t) nec765->secSize)  {
              /* Read error */
              uiErrPrintf("NEC765: Image file read error - %s\n", strerror(errno));
              return(-1);
          }
          nec765->sectorIdx = 0;
          break;

      case CMD_WRITE_DATA :
          cnt = fwrite(nec765->sectorBuffer, 1, nec765->secSize, nec765->imgFile);
          if (cnt != (size_t) nec765->secSize)  {
              /* Write error */
              uiErrPrintf("NEC765: Image file write error - %s\n", strerror(errno));
              return(-1);
          }
          break;

      default :
		  uiErrPrintf("NEC765: Internal error - Invalid mode for trasferSector()\n");
          return(-1);

    }

    return (0);
}



/*
 * c m d C h e c k
 *
 * Check command code and return number of command bytes
 */
static int cmdCheck(NEC765 *nec765, u_char cmd)
{
    int  cnt;

    switch (cmd & 0x1F)  {
        case CMD_SPECIFY :   cnt = 3; break;
        case CMD_SEEK :      cnt = 3; break;
        case CMD_RECAL :     cnt = 2; break;
        case CMD_READ_DATA : cnt = 9; break;
        case CMD_WRITE_DATA: cnt = 9; break;
        case CMD_SENSE_INT : cnt = 1; break;
        case CMD_FORMAT_TRK: cnt = 6; break;
        default :            cnt = 1; break;
    }

    return(cnt);
}



/*
 * s e t I n t e r r u p t
 *
 * Request Interrupt
 */
static void setInterrupt(Device *dev)
{
}



/*
 * c m d E x e c u t e
 *
 * Execute command code
 */
static int cmdExecute(Device *dev, NEC765 *nec765)
{
    
    switch (nec765->cmd[0] & 0x1F)  {
      case CMD_SPECIFY :
        nec765->phase = PH_IDLE;
        nec765->msr = MSR_RQM;
        break;

      case CMD_SEEK :
        nec765->curTrack = nec765->cmd[2];
        nec765->phase = PH_IDLE;
        nec765->msr = MSR_RQM;
        intSetIRQ(dev);
        break;

      case CMD_RECAL :
        nec765->curTrack = 0;
        nec765->phase = PH_IDLE;
        nec765->msr = MSR_RQM;
        intSetIRQ(dev);
        break;

      case CMD_SENSE_INT :
        nec765->stat[0] = 0x20;
        nec765->stat[1] = nec765->curTrack;
        nec765->sidx = 0; nec765->scnt = 2;
        nec765->phase = PH_RESULT;
        nec765->msr   = (MSR_RQM | MSR_DIO | MSR_CIP);
        break;

      case CMD_READ_DATA :
        nec765->curHead  = (nec765->cmd[1] & 0x04) ? 1 : 0;
        nec765->secSize = (128 << nec765->cmd[5]);
        nec765->curTrack = nec765->cmd[2];
        nec765->curSector = nec765->cmd[4];
        transferSector(nec765, CMD_READ_DATA);
        nec765->phase = PH_EXECUTION;
        nec765->msr   = (MSR_RQM | MSR_DIO | MSR_EXC | MSR_CIP);
        break;

      case CMD_WRITE_DATA:
        nec765->curHead  = (nec765->cmd[1] & 0x04) ? 1 : 0;
        nec765->secSize = (128 << nec765->cmd[5]);
        nec765->curTrack = nec765->cmd[2];
        nec765->curSector = nec765->cmd[4];
        nec765->sectorIdx = 0;
        nec765->phase = PH_EXECUTION;
        nec765->msr   = (MSR_RQM | MSR_EXC | MSR_CIP);
        break;

      case CMD_FORMAT_TRK :
        nec765->curHead  = (nec765->cmd[1] & 0x04) ? 1 : 0;
        nec765->secSize  = (128 << nec765->cmd[2]);
        nec765->secTrack = nec765->cmd[3];
        nec765->dataPat  = nec765->cmd[5];
        nec765->secCount = 0;
        nec765->idIdx = 0;
        nec765->phase = PH_EXECUTION;
        nec765->msr   = (MSR_RQM | MSR_EXC | MSR_CIP);
        break;

      default :
        nec765->stat[0] = 0x80;
        nec765->sidx = 0;
        nec765->scnt = 1;
        nec765->phase = PH_RESULT;
        nec765->msr = (MSR_RQM | MSR_DIO);
        break;
    }

    return(0);
}



/*
 * n e c 7 6 5 R e s e t
 *
 * Reset the device
 */
int nec765Reset(Device *dev)
{
    NEC765  *nec765 = (NEC765 *) dev->priv;

    if (!quiet)  uiPrintf(" NEC765");

    nec765->numOfHeads = 2;

    nec765->msr = 0x00;
    nec765->dsr = 0x02;

    nec765->secTrack = 16;
    nec765->secSize = 256;

    nec765->phase = PH_IDLE;
    nec765->msr = MSR_RQM;

    return 0;
}



/*
 * n e c 7 6 5 R e a d
 *
 * Read one byte
 */
static u_long nec765Read(Device *dev, u_long offset)
{
    NEC765  *nec765 = (NEC765 *) dev->priv;
    u_char  data = 0xFF;


    switch (offset & (REGSIZE-1))  {
      case MSR :
          data = nec765->msr;
          break;

      case DATA :
          switch (nec765->phase)  {
            case PH_IDLE :
                data = 0xFF;
                break;

            case PH_COMMAND :
                data = 0xFE;
                break;

            case PH_EXECUTION :
                if ((nec765->cmd[0] & 0x1F) == CMD_READ_DATA)  {
                    data = nec765->sectorBuffer[nec765->sectorIdx];
                    if (++nec765->sectorIdx >= nec765->secSize) {
                        /* End of read command */
                        nec765->stat[0] = 0x00;
                        nec765->stat[1] = 0x00;
                        nec765->stat[2] = 0x00;
                        nec765->stat[3] = nec765->cmd[2];
                        nec765->stat[4] = nec765->cmd[3];
                        nec765->stat[5] = nec765->cmd[4];
                        nec765->stat[6] = nec765->cmd[5];
                        nec765->sidx = 0; nec765->scnt = 7;
                        nec765->phase = PH_RESULT;
                        nec765->msr   = (MSR_RQM | MSR_DIO | MSR_CIP);
                    }
                    else  {
                        nec765->phase = PH_EXECUTION;
                        nec765->msr   = (MSR_RQM | MSR_DIO | MSR_EXC | MSR_CIP);
                    }
                }
                break;

            case PH_RESULT :
                data = nec765->stat[nec765->sidx++];
                if (nec765->sidx >= nec765->scnt) {
                    nec765->phase = PH_IDLE;
                    nec765->msr   = MSR_RQM;
                }
                else  {
                    nec765->phase = PH_RESULT;
                    nec765->msr   = (MSR_RQM | MSR_DIO | MSR_CIP);
                }
                break;
          }
          break;

      default :
          data = 0xFF;
          break;
    }

    return (u_long) data;
}



/*
 * n e c 7 6 5 W r i t e
 *
 * Write one byte
 */
static void nec765Write(Device *dev, u_long offset, u_char data)
{
    NEC765  *nec765 = (NEC765 *) dev->priv;
    int     i;

    switch (offset & (REGSIZE-1))  {
      case DATA :
          switch (nec765->phase)  {
            case PH_IDLE :
                nec765->cmd[0] = data;
                nec765->cidx = 1;
                nec765->ccnt = cmdCheck(nec765, data);
                if (nec765->cidx < nec765->ccnt) {
                    nec765->phase = PH_COMMAND;
                    nec765->msr   = (MSR_RQM | MSR_CIP);
                }
                else {
                    nec765->phase = PH_EXECUTION;
                    cmdExecute(dev, nec765);
                }
                break;

            case PH_COMMAND :
                nec765->cmd[nec765->cidx++] = data;
                if (nec765->cidx < nec765->ccnt)  {
                    nec765->phase = PH_COMMAND;
                    nec765->msr   = (MSR_RQM | MSR_CIP);
                }
                else  {
                    nec765->phase = PH_EXECUTION;
                    cmdExecute(dev, nec765);
                }
                break;

            case PH_EXECUTION :
                if ((nec765->cmd[0] & 0x1F) == CMD_WRITE_DATA)  {
                    nec765->sectorBuffer[nec765->sectorIdx] = data;
                    if (++nec765->sectorIdx >= nec765->secSize)  {
                        /* Buffer filled. Transfer to disk */
                        transferSector(nec765, CMD_WRITE_DATA);
                        nec765->stat[0] = 0x00;
                        nec765->stat[1] = 0x00;
                        nec765->stat[2] = 0x00;
                        nec765->stat[3] = nec765->cmd[2];
                        nec765->stat[4] = nec765->cmd[3];
                        nec765->stat[5] = nec765->cmd[4];
                        nec765->stat[6] = nec765->cmd[5];
                        nec765->sidx = 0; nec765->scnt = 7;
                        nec765->phase = PH_RESULT;
                        nec765->msr   = (MSR_RQM | MSR_DIO | MSR_CIP);
                    }
                    else  {
                        nec765->phase = PH_EXECUTION;
                        nec765->msr   = (MSR_RQM | MSR_EXC | MSR_CIP);
                    }
                } else if ((nec765->cmd[0] & 0x1F) == CMD_FORMAT_TRK)  {
                    nec765->idBuffer[nec765->idIdx] = data;
                    if (++nec765->idIdx >= 4)  {
                        for (i = 0; i < nec765->secSize; i++)
                            nec765->sectorBuffer[i] = nec765->cmd[5];
                        nec765->curSector = nec765->secCount+1;
                        transferSector(nec765, CMD_WRITE_DATA);
                        if (++nec765->secCount >= nec765->secTrack)  {
                            nec765->stat[0] = 0x00;
                            nec765->stat[1] = 0x00;
                            nec765->stat[2] = 0x00;
                            nec765->stat[3] = nec765->cmd[2];
                            nec765->stat[4] = nec765->cmd[3];
                            nec765->stat[5] = nec765->cmd[4];
                            nec765->stat[6] = nec765->cmd[5];
                            nec765->sidx = 0; nec765->scnt = 7;
                            nec765->phase = PH_RESULT;
                            nec765->msr   = (MSR_RQM | MSR_DIO | MSR_CIP);
                        }
                        else {
                            nec765->idIdx = 0;
                            nec765->phase = PH_EXECUTION;
                            nec765->msr   = (MSR_RQM | MSR_EXC | MSR_CIP);
                        }
                    }
                    else  {
                        nec765->phase = PH_EXECUTION;
                        nec765->msr   = (MSR_RQM | MSR_EXC | MSR_CIP);
                    }
                }
                break;
 
            case PH_RESULT :
                nec765->phase = PH_IDLE;
                nec765->msr = MSR_RQM;
                break;
          }
          break;

      default:
          break;
    }
}



/*
 * n e c 7 6 5 I A c k
 *
 * Interrupt Acknowledge
 */
static u_char nec765IAck(Device *dev, int level)
{
    intClearIRQ(dev);

    return 0x44;
}



/*
 * i n i t N E C 7 6 5
 *
 * Initialize the device
 */
int initNEC765(Device *dev, u_long base, u_long size, void *arg)
{
    NEC765   *nec765;
    char     *fileName = (char *)arg;
    FILE     *imgFile;
    struct stat statBuffer;

    if ((nec765 = malloc(sizeof(NEC765))) == NULL)  {
        uiErrPrintf("NEC765: Malloc of %ld bytes failed - %s\n", size, strerror(errno));
        return(errno);
    }

    /* Open image file */
    if ((imgFile = fopen(fileName, "r+b")) == NULL)  {
        uiErrPrintf("NEC765: Can't open %s - %s\n", fileName, strerror(errno));
        return(errno);
    }

    if (stat(fileName, &statBuffer) == -1)  {
        uiErrPrintf("NEC765: Can't get size of %s - %s\n", fileName, strerror(errno));
        return(errno);
    }

    nec765->imgFile = imgFile;
    nec765->diskSize = statBuffer.st_size / 512;

    /* Setup device structure */
    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (u_char *) nec765;
    dev->devReset = nec765Reset;
    dev->devRead  = nec765Read;
    dev->devWrite = nec765Write;
    dev->devIAck  = nec765IAck;
    dev->devCycle = NULL;

    return 0;
}
