/*
 * Most of this code was written by Marwan Shaban at North Carolina
 * University's Electrical and Computer Engineering department in 1989.
 *
 * Marwan Shaban, shaban@ece36h.ncsu.edu
 *
 * Ported to Linux as a part of the 68000 system simulator by
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.3 - May 02 2004 20:59:11
 *
 * 1  21.04.2004  
 * 2  01.05.2004  Modification in ROR and ROL
 * 3  02.05.2004  Same modification in ASR, ROXL, ROXR
 */

#include <stdio.h>
#include <sys/types.h>

#include "s68.h"
#include "ui.h"
#include "m68000.h"
#include "decoder.h"
#include "opcodes.h"



extern int   cpuState;          /* CPU state */
extern long  inst;              /* Current instruction */

/* Return of effAddr() */
static long  *EA1, *EA2;        /* Pointer to register memaddr or immdata */
static long  memaddr;           /* hold the memory address operand of EA */
static long  immdata;           /* hold an immediate data operand of EA */



/*
 * a R e g
 *
 * Return address register index
 */
static int aReg(int areg)
{
    if (areg != 7)
        return areg;
    else
        return (SR & 0x2000) ? 8 : 7;
}



/*
 * s i g n E x t e n d
 *
 * sign-extends a number from byte or word to long
 */
static void signExtend(int number, long size, long *result)
{

    if (size == SZ_WORD)  {
        if (number & 0x8000)
            *result = number | 0xffff0000;
        else
            *result = number & 0x0000ffff;
    }
    else if (size == SZ_BYTE)  {
        if (number & 0x80)
            *result = number | 0xffffff00;
        else
            *result = number & 0x000000ff;
    }
    else  {
        *result = number & 0xffffffff;
    }
}



/*
 * g e t
 *
 * Returns the value of the location referenced regardless of whether it is
 * a memory location or a 68000 register location.
 */
static void get(long *EA, long *value, long size)
{
    if (EA == &memaddr)  {
        memGet(*EA, size, value);
    }
    else  {
        switch (size)  {
          case SZ_LONG : *value = *EA & 0xffffffff; break;
          case SZ_WORD : *value = *EA & 0x0000ffff; break;
          case SZ_BYTE : *value = *EA & 0x000000ff; break;
        }
    }
}



/*
 * p u t
 *
 * Performs the task of putting the result of some operation into a destination
 * location "according to size".  This means that the bits of the destination
 * that are in excess to the size of the operation are not affected.
 * This function provides a general-purpose mechanism for putting the result of
 * an operation in the destination, no matter whether the destination is
 * a memory location or a 68000 register.  The data is placed in the destination
 * correctly and the rest of the bits in a register are left alone.
 */
static void put(long *dst, long src, long size)
{
    if (dst == &memaddr)
        memPut(src, *dst, size);
    else  {
        switch (size)  {
          case SZ_BYTE :
            *dst = (src & 0x000000ff) | (*dst & ~0x000000ff);
			break;
          case SZ_WORD :
            *dst = (src & 0x0000ffff) | (*dst & ~0x0000ffff);
			break;
          case SZ_LONG :
            *dst = (src & 0xffffffff) | (*dst & ~0xffffffff);
			break;
        }
    }
}



/*
 * D e c o d e S i z e
 *
 * Decodes the size field in the instruction being processed
 * and returns a mask to be used in instruction execution.
 */
static int decodeSize(long *result)
{
    int ret = SUCCESS;

    switch ((inst >> 6) & 0x0003)  {
      case 0 : *result = SZ_BYTE; TrcPrintf((".b\t")); break;
      case 1 : *result = SZ_WORD; TrcPrintf((".w\t")); break;
      case 2 : *result = SZ_LONG; TrcPrintf((".l\t")); break;
      default: *result = 0; ret = ILLEGAL_INST; TrcPrintf(("\t")); break;
    }

    return ret;
}


/*
 * e f f A d d r
 *
 * Decodes the effective address field in the current instruction,
 * returns a pointer to the effective address either in EA1 or EA2
 */
static int effAddr(long size, int mask)
{
    int   mode, reg, addr;
    int   move; 
    int   legal = 0;
    long  ext, inc_size, ind_reg, disp;
    long  *value = NULL;


    if (((inst >= 0x1000) && (inst <= 0x3FFF)) && (mask == DATA_ALT_ADDR))  {
        move = 1;
        /* reg and mode are reversed in MOVE dest EA */
        addr = (inst >> 6) & 0x003f;
        mode = (addr & REG_MASK);
        reg  = (addr & MODE_MASK) >> 3;
    }
    else  {
        move = 0;
        addr = inst & 0x003f;
        mode = (addr & MODE_MASK) >> 3;
        reg  = (addr & REG_MASK);
    }

    switch (mode) {
        case 0 :
            if (mask & bit_1)  {
                value = &D[reg];
                TrcPrintf(("D%d", reg));
                legal = 1;
            }
            break;
        case 1 :
            if (mask & bit_2)  {
                value = &A[aReg(reg)];
                TrcPrintf(("A%d", reg));
                legal = 1;
            }
            break;
        case 2 :
            if (mask & bit_3)  {
                memaddr = A[aReg(reg)];
                value = &memaddr;
                TrcPrintf(("(A%d)", reg));
                legal = 1;
            }
            break;
        case 3 :
            if (mask & bit_4)  {
                inc_size = (size == SZ_BYTE) ? 1 : ((size == SZ_WORD) ? 2 : 4);
                memaddr = A[aReg(reg)];
                value = &memaddr;
                A[aReg(reg)] += inc_size;
                TrcPrintf(("(A%d)+", reg));
                legal = 1;
            }
            break;
        case 4 :
            if (mask & bit_5)  {
                inc_size = (size == SZ_BYTE) ? 1 : ((size == SZ_WORD) ? 2 : 4);
                A[aReg(reg)] -= inc_size;
                memaddr = A[aReg(reg)];
                value = &memaddr;
                TrcPrintf(("-(A%d)", reg));
                legal = 1;
            }
            break;
        case 5 :
            if (mask & bit_6)  {
                memRequest(&PC, SZ_WORD, &ext);
                signExtend(ext, SZ_WORD, &ext);
                memaddr = A[aReg(reg)] + ext;
                value = &memaddr;
                TrcPrintf(("%04x(A%d)", (unsigned short) ext, reg));
                legal = 1;
            }
            break;
        case 6 :
            if (mask & bit_7)  {
                /* Get extension word */
                memRequest(&PC, SZ_WORD, &ext);
                disp = ext & 0x00ff;
                signExtend(disp, SZ_BYTE, &disp);
                /* Get index register value */
                if (ext & 0x8000)
                    ind_reg = A[aReg((ext & 0x7000) >> 12)];
                else
                    ind_reg = D[(ext & 0x7000) >> 12];
                /* Get correct length for index register */
                if (!(ext & 0x0800)) {
                    signExtend(ind_reg, SZ_WORD, &ind_reg);
                }
                memaddr = A[aReg(reg)] + disp + ind_reg;
                value = &memaddr;
                TrcPrintf(("%02x(A%d,%c%d.%c)", (unsigned char) disp, reg, (ext & 0x8000)?'A':'D', (int)((ext & 0x7000) >> 12), (ext & 0x0800) ? 'l' : 'w'));
                legal = 1;
            }
            break;
        case 7 :
            switch (reg) {
                case 0 :
                    if (mask & bit_8) {
                        memRequest(&PC, SZ_WORD, &ext);
                        signExtend(ext, SZ_WORD, &ext);
                        memaddr = ext;
                        value = &memaddr;
                        TrcPrintf(("%08lx.w", ext));
                        legal = 1;
                    }
                    break;
                case 1 :
                    if (mask & bit_9)  {
                        memRequest(&PC, SZ_LONG, &ext);
                        memaddr = ext;
                        value = &memaddr;
                        TrcPrintf(("%08lx.l", ext));
                        legal = 1;
                    }
                    break;
                case 2 :
                    if (mask & bit_10) {
                        memRequest(&PC, SZ_WORD, &ext);
                        signExtend(ext, SZ_WORD, &ext);
                        TrcPrintf(("%04x(PC)", (unsigned short) ext));
                        memaddr = PC - 2  + ext;
                        value = &memaddr;
                        legal = 1;
                    }
                    break;
                case 3 :
                    if (mask & bit_11)  {
                        /* Get extension word */
                        memRequest(&PC, SZ_WORD, &ext);
                        disp = ext & 0x00ff;
                        signExtend(disp, SZ_BYTE, &disp);
                        /* Get index register value */
                        if (ext & 0x8000)
                            ind_reg = A[aReg((ext & 0x7000) >> 12)];
                        else
                            ind_reg = D[(ext & 0x7000) >> 12];
                        /* Get correct length for index register */
                        if (!(ext & 0x0800))
                            signExtend(ind_reg, SZ_WORD, &ind_reg);
                        ext = ext & 0x00ff;
                        memaddr = PC - 2 + disp + ind_reg;
                        value = &memaddr;
                        TrcPrintf(("%02x(PC,%c%d.%c)", (unsigned char) disp, (ext & 0x8000)?'A':'D', (int)((ext & 0x7000) >> 12), (ext & 0x0800) ? 'l' : 'w'));
                        legal = 1;
                    }
                    break;
                case 4 :
                    if (mask & bit_12)  {
                        if ((size == SZ_BYTE) || (size == SZ_WORD))
                            memRequest(&PC, SZ_WORD, &ext);
                        else
                            memRequest(&PC, SZ_LONG, &ext);
                        immdata = ext;
                        value = &immdata;
                        TrcPrintf(("#%lx", ext));
                        legal = 1;
                    }
                    break;
            }
            break;
    }
    if (!legal)  return ILLEGAL_INST;

    if (move)  {
        EA2 = value;
    }
    else {
        EA1 = value;
    }

    return SUCCESS;
}



/*
 * c c r U p d a t e
 *
 * Updates the five condition codes according to the codes passed as parameters.
 * Each of the condition codes has a number of ways it can be calculated, and
 * the appropriate method of computation is passed as the parameter for that
 * condition code.
 * The src, dst, and res operands contain the source, destination, and result
 * of the instruction requesting updating the condition codes.
 * Also, for shift and rotate instructions the shift count needs to be passed.
 */
static int ccrUpdate(int x, int n, int z, int v, int c, long source, long dest, long result, long size, int r)
{
    int   x_bit, n_bit, z_bit, v_bit, c_bit;
    long  Rm, Dm, Sm;
    long  m;


    if (size == SZ_BYTE)  {
        source &= 0x000000ff;
        dest   &= 0x000000ff;
        result &= 0x000000ff;
        m = 7;
        Sm = source & 0x0080;
        Rm = result & 0x0080;
        Dm = dest & 0x0080;
    }
    else if (size == SZ_WORD)  {
        source &= 0x0000ffff;
        dest   &= 0x0000ffff;
        result &= 0x0000ffff;
        m = 15;
        Sm = source & 0x8000;
        Rm = result & 0x8000;
        Dm = dest & 0x8000;
    }
    else  {
        source &= 0xffffffff;
        dest   &= 0xffffffff;
        result &= 0xffffffff;
        m = 31;
        Sm = source & 0x80000000;
        Rm = result & 0x80000000;
        Dm = dest & 0x80000000;
    }

    /*
     * Calculate each of the five condition codes according to the requested
     * method of calculation
     */
    switch (n) {
      case GEN :
        n_bit = (Rm) ? nbit : 0;
        break;
      default :
        n_bit = SR & nbit;
        break;
    }

    switch (z) {
      case GEN :
        z_bit = (result == 0) ? zbit : 0;
        break;
      case CASE_1 :
        z_bit = ((SR & zbit) && (result == 0)) ? zbit : 0;
        break;
      default :
        z_bit = SR & zbit;
        break;
    }

    switch (v) {
      case ZER :
        v_bit = 0;
        break;
      case CASE_1 :
        v_bit = ((Sm && Dm && !Rm) || (!Sm && !Dm && Rm)) ? vbit : 0;
        break;
      case CASE_2 :
        v_bit = ((!Sm && Dm && !Rm) || (Sm && !Dm && Rm)) ?  vbit : 0;
        break;
      case CASE_3 :
        v_bit = (Dm && Rm) ? vbit : 0;
        break;
      default :
        v_bit = SR & vbit;
        break;
    }

    switch (c) {
      case ZER :
        c_bit = 0;
        break;
      case CASE_1 :
        c_bit = (SR & xbit) ? cbit : 0;
        break;
      case CASE_2 :
        c_bit = ((dest >> (r-1)) & 1) ? cbit : 0;
        break;
      case CASE_3 :
        c_bit = ((dest >> (m-r+1)) & 1) ? cbit : 0;
        break;
      case CASE_4 :
        c_bit = (Dm || Rm) ? cbit : 0;
        break;
      case CASE_5 :
        c_bit = ((Sm && Dm) || (!Rm && Dm) || (Sm && !Rm)) ?  cbit : 0;
        break;
      case CASE_6 :
        c_bit = ((Sm && !Dm) || (Rm && !Dm) || (Sm && Rm)) ?  cbit : 0;
        break;
      default :
        c_bit = SR & cbit;
        break;
    }

    switch (x) {
      case GEN :
        x_bit = c_bit ? xbit : 0;
        break;
      default :
        x_bit = SR & xbit;
        break;
    }

    /* set SR according to results */
    SR = SR & 0xffe0;
    SR = SR | (x_bit | n_bit | z_bit | v_bit | c_bit);

    return SUCCESS;
}



/*
 * c h e c k C o n d i t i o n
 *
 * Checks for the truth of a certain condition and returns the result.
 */
int checkCondition(int condition)
{
    int   result = 0;

    switch (condition)  {
      case T :
        TrcPrintf(("t"));
        result = 1;                                   /* true */
        break;

      case F :
        TrcPrintf(("f"));
        result = 0;                                   /* false */
        break;

      case HI :
        TrcPrintf(("hi"));
        result = !(SR & cbit) && !(SR & zbit);        /* high */
        break;

      case LS :
        TrcPrintf(("ls"));
        result = (SR & cbit) || (SR & zbit);          /* low or same */
        break;

      case CC :
        TrcPrintf(("cc"));
        result = !(SR & cbit);                        /* carry clear */
        break;

      case CS :
        TrcPrintf(("cs"));
        result = (SR & cbit);                         /* carry set */
        break;

      case NE :
        TrcPrintf(("ne"));
        result = !(SR & zbit);                        /* not equal */
        break;

      case EQ :
        TrcPrintf(("eq"));
        result = (SR & zbit);                         /* equal */
        break;

      case VC :
        TrcPrintf(("vc"));
        result = !(SR & vbit);                        /* overflow clear */
        break;

      case VS :
        TrcPrintf(("vs"));
        result = (SR & vbit);                         /* overflow set */
        break;

      case PL :
        TrcPrintf(("pl"));
        result = !(SR & nbit);                        /* plus */
        break;

      case MI :
        TrcPrintf(("mi"));
        result = (SR & nbit);                         /* minus */
        break;

      case GE :
        TrcPrintf(("ge"));
        result = ((SR & nbit) && (SR & vbit)) || (!(SR & nbit) && !(SR & vbit));
        break;

      case LT :
        TrcPrintf(("lt"));
        result = ((SR & nbit) && !(SR & vbit)) || (!(SR & nbit) && (SR & vbit));
        break;

      case GT :
        TrcPrintf(("gt"));
        result = ((SR & nbit) && (SR & vbit) && !(SR & zbit)) ||
                  (!(SR & nbit) && !(SR & vbit) && !(SR & zbit));
        break;

      case LE :
        TrcPrintf(("le"));
        result = ((SR & nbit) && !(SR & vbit)) ||
                  (!(SR & nbit) && (SR & vbit)) || (SR & zbit);
        break;
    }

    return result;
}



int MOVE(void)
{
    long  size;
    long  src, dst;

    switch ((inst >> 12) & 0x03)  {
        case 0x01 : size = SZ_BYTE; break;         /* bit pattern '01' */
        case 0x03 : size = SZ_WORD; break;         /* bit pattern '10' */
        case 0x02 : size = SZ_LONG; break;         /* bit pattern '11' */
        default   : return ILLEGAL_INST;        /* bad size field */
    }

    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;
    get(EA1, &src, size);
    TrcPrintf((","));
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    dst = src;
    put(EA2, dst, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, dst, size, 0);

    return SUCCESS;
}


int MOVEP(void)
{
    int   address, Dx, disp, count, direction, reg;
    long  temp;

    memRequest(&PC, SZ_WORD, &temp);
    signExtend(temp, SZ_WORD, &temp);

    address = A[aReg(inst & 0x0007)] + temp;
    direction = inst & 0x80;

    reg = (inst >> 9) & 0x07;
    if (inst & 0x40)  {
        Dx = D[reg] & 0xffffffff;
        count = 4;
    }
    else  {
        Dx = D[reg] & 0x0000ffff;
        count = 2;
    }

    for (;count > 0; count--)  {
        disp = 8 * (count - 1);
        if (direction)  {
            memPut((long)((Dx >> disp) & 0x000000ff) , address, SZ_BYTE);
        }
        else  {
            memGet(address, SZ_BYTE, &temp);
            switch (count) {
              case 4 :
                D[reg] = (D[reg] & 0x00ffffff) | (temp << 24);
                break;
              case 3 :
                D[reg] = (D[reg] & 0xff00ffff) | (temp << 16);
                break;
              case 2 :
                D[reg] = (D[reg] & 0xffff00ff) | (temp << 8);
                break;
              case 1 :
                D[reg] = (D[reg] & 0xffffff00) | (temp << 0);
                break;
            }
        }
        address += 2;
    }

    return SUCCESS;
}


int MOVEA(void)
{
    int   reg;
    long  size, src;

    size = (inst & 0x1000) ? SZ_WORD : SZ_LONG;
    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;

    get(EA1, &src, size);
    if (size == SZ_WORD)  signExtend(src, SZ_WORD, &src);

    reg = (inst >> 9) & 0x07;
    A[aReg(reg)] = src;

    TrcPrintf((",A%d", reg));

    return SUCCESS;
}


int MOVE_SR_(void)
{
    TrcPrintf(("SR,"));

    if (effAddr(SZ_WORD, DATA_ALT_ADDR))  return ILLEGAL_INST;

    put(EA1, SR, SZ_WORD);

    return SUCCESS;
}


int MOVE_CCR(void)
{
    long  temp;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;

    get(EA1, &temp, SZ_WORD);
    put(&SR, temp, SZ_BYTE);

    TrcPrintf((",CCR"));

    return SUCCESS;
}


int MOVE_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;

    get(EA1, &temp, SZ_WORD);
    put(&SR, temp, SZ_WORD);

    TrcPrintf((",SR"));

    return SUCCESS;
}


int MOVEM(void)
{
    int   mem2reg;
    long  inc;
    long  address;
    long  size, maskList, temp;
    int   i;

    memRequest(&PC, SZ_WORD, &maskList);

    size = (inst & 0x0040) ? SZ_LONG : SZ_WORD;
    inc  = (size == SZ_WORD) ? 2 : 4;

    TrcPrintf(((inst & 0x0040) ? ".l\t" : ".w\t"));

    mem2reg = (inst & 0x0400);
    if (mem2reg)  {
        if (effAddr(size, (CONTROL_ADDR | bit_4)))  return ILLEGAL_INST;
        if (trace)  {
            int  f = 0;

            TrcPrintf((","));
            for (i = 0; i < 16; i++)  {
              if (maskList & (1 << i))  {
                if (f++)  TrcPrintf(("/"));
                TrcPrintf(((i < 8) ? "D%d":"A%d", i%8));
              }
            }
        }
    }
    else  {
        if (trace)  {
            int  f = 0;

            if ((inst & 0x0038) == 0x0020) {
              /* predecrement mode */
              for (i = 15; i > 0; i--)
                if (maskList & (1 << i))  {
                  if (f++)  TrcPrintf(("/"));
                  TrcPrintf(((i > 7) ? "D%d":"A%d", (15-i)%8));
                }
            }
            else  {
              for (i = 0; i < 16; i++)
                if (maskList & (1 << i))  {
                  if (f++)  TrcPrintf(("/"));
                  TrcPrintf(((i < 8) ? "D%d":"A%d", i%8));
                }
            }
            TrcPrintf((","));
        }
        if (effAddr(size, (CONT_ALT_ADDR | bit_5)))  return ILLEGAL_INST;
    }
    address = *EA1;

    if (mem2reg) {
        /* Move memory to register */
        for (i = 0; i < 16; i++)  {
            if (maskList & (1 << i)) {
                memGet(address, size, &temp);
                if (size == SZ_WORD)  signExtend(temp, SZ_LONG, &temp);
                if (i < 8)
                    D[i] = temp;
                else
                    A[aReg(i - 8)] = temp;
                address += inc;
            }
        }
    }
    else  {
        /* Move register to memory */
        if ((inst & 0x0038) == 0x0020) {
            /* Predecrement mode */
            for (i = 0; i < 16; i++)  {
                if (maskList & (1 << i)) {
                    if (i < 8)
                        memPut(A[aReg(7-i)], address, size);
                    else
                        memPut(D[15 - i], address, size);
                    address -= inc;
                }
            }
        }
        else  {
            for (i = 0; i < 16; i++)  {
                if (maskList & (1 << i)) {
                    if (i < 8)
                        memPut(D[i], address, size);
                    else
                        memPut(A[aReg(i - 8)], address, size);
                    address += inc;
                }
            }
        }
    }

    if ((inst & 0x0038) == 0x0020)  {
        A[aReg(inst & 0x7)] = address + inc;
    }
    else if ((inst & 0x0038) == 0x0018)  {
        A[aReg(inst & 0x7)] = address;
    }

    return SUCCESS;
}


int MOVE_USP(void)
{
    int   regNum;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    regNum = aReg(inst & 0x0007);

    if (inst & 0x0008) {
        A[regNum] = A[7];
        TrcPrintf(("USP,a%d", regNum));
    }
    else  {
        A[7] = A[regNum];
        TrcPrintf(("a%d,USP", regNum));
    }

    return SUCCESS;
}


int MOVEQ(void)
{
    int   reg;
    long  src, dst;

    reg = (inst >> 9) & 0x7;
    src = inst & 0x00ff;
    signExtend(src, SZ_BYTE, &src);

    D[reg] = dst = src;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, D[reg], SZ_LONG, 0);

    TrcPrintf(("\t#%ld,D%d", src, reg));

    return SUCCESS;
}


int EXG(void)
{
    long  temp_reg;
    int   Rx, Ry;

    Rx = (inst >> 9) & 0x07;
    Ry = inst & 0x0007;

    switch ((inst >> 3) & 0x1f) {
      case 0x08:
        temp_reg = D[Rx];
        D[Rx] = D[Ry];
        D[Ry] = temp_reg;
        TrcPrintf(("D%d,D%d", Rx, Ry));
        break;
      case 0x09:
        temp_reg = A[aReg(Rx)];
        A[aReg(Rx)] = A[aReg(Ry)];
        A[aReg(Ry)] = temp_reg;
        TrcPrintf(("A%d,A%d", Rx, Ry));
        break;
      case 0x11:
        temp_reg = D[Rx];
        D[Rx] = A[aReg(Ry)];
        A[aReg(Ry)] = temp_reg;
        TrcPrintf(("D%d,A%d", Rx, Ry));
        break;
      default :
        return ILLEGAL_INST;
        break;
    }

    return SUCCESS;
}


int LEA(void)
{
    int   reg;

    reg = (inst >> 9) & 0x07;

    if (effAddr(SZ_LONG, CONTROL_ADDR))  return ILLEGAL_INST;
    A[aReg(reg)] = *EA1;

    TrcPrintf((",A%d", reg));

    return SUCCESS;
}


int PEA(void)
{
    if (effAddr(SZ_LONG, CONTROL_ADDR))  return ILLEGAL_INST;

    A[aReg(7)] -= 4;
    memPut(*EA1 , A[aReg(7)], SZ_LONG);

    return SUCCESS;
}


int LINK(void)
{
    int   reg;
    long  temp, disp;

    reg = inst & 0x0007;

    memRequest(&PC, SZ_WORD, &temp);
    signExtend(temp, SZ_WORD, &disp);

    A[aReg(7)] -= 4;
    memPut(A[reg], A[aReg(7)], SZ_LONG);
    A[reg] = A[aReg(7)];
    A[aReg(7)] = A[aReg(7)] + disp;

    TrcPrintf(("#%04x,A%d", (unsigned short) temp, reg));

    return SUCCESS;
}


int UNLK(void)
{
    int   reg;

    reg = inst & 0x0007;

    A[aReg(7)] = A[reg];
    memGet(A[aReg(7)], SZ_LONG, &A[reg]);
    A[aReg(7)] += 4;

    TrcPrintf(("A%d", reg));

    return SUCCESS;
}


int ADD(void)
{
    int   addr_modes_mask;
    int   reg;
    long  size;
    long  src, dst, res;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : ALL_ADDR;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, addr_modes_mask))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        src = D[reg];
        get(EA1, &dst, size);
        res = src + dst;
        put(EA1, res, size);
    }
    else {
        get(EA1, &src, size);
        dst = D[reg];
        res = src + dst;
        put(&D[reg], res, size);
    }
    ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, src, dst, res, size, 0);

    return SUCCESS;
}


int ADDA(void)
{
    int   reg;
    long  size;
    long  src, dst;

    size = (inst & 0x0100) ? SZ_LONG : SZ_WORD;
    TrcPrintf((".%c\t", (size == SZ_LONG) ? 'l' : 'w'));

    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;

    get(EA1, &src, size);
    dst = A[aReg(reg)];
    put(&A[aReg(reg)], src + dst, size);

    TrcPrintf((",A%d", reg));

    return SUCCESS;
}


int ADDI(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;

    memRequest(&PC, size, &src);

    TrcPrintf(("#%lx,", src));

    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    res = src + dst;
    put(EA1, res, size);
    ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, src, dst, res, size, 0);

    return SUCCESS;
}


int ADDQ(void)
{
    long  size, src, dst, res;


    if (decodeSize(&size))  return ILLEGAL_INST;

    if ((inst & 0x0038) == 0x0008)
        /* if address reg direct, operation is long */
        size = SZ_LONG;

    src = ((inst >> 9) & 0x07);
    if (src == 0)  src = 8;

    TrcPrintf(("#%ld,", src));

    if (effAddr(size, ALTERABLE_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    res = src + dst;
    put(EA1, res, size);

    if (!((inst & 0x0038) == 0x0008))
        /* Address reg direct: cc's not affected */
        ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, src, dst, res, size, 0);

    return SUCCESS;
}


int ADDX(void)
{
    long  size, src, dst, res;
    int   Rx, Ry;

    if (decodeSize(&size))  return ILLEGAL_INST;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        Rx = aReg(Rx);
        Ry = aReg(Ry);
        if (size == SZ_BYTE) {
            A[Rx]--;
            A[Ry]--;
        }
        else if (size == SZ_WORD) {
            A[Rx] -= 2;
            A[Ry] -= 2;
        }
        else {
            A[Rx] -= 4;
            A[Ry] -= 4;
        }
        memGet(A[Ry], size, &src);
        memGet(A[Rx], size, &dst);
        memPut(A[aReg(Rx)], src + dst + ((SR & xbit) >> 4), size);
        memGet(A[Rx], size, &res);
    }
    else  {
        src = D[Ry] & size;
        dst = D[Rx] & size;
        put(&D[Rx], src + dst + ((SR & xbit) >> 4), size);
        res = D[Rx] & size;
    }
    ccrUpdate(GEN, GEN, CASE_1, CASE_1, CASE_5, src, dst, res, size, 0);

    return SUCCESS;
}


int SUB(void)
{
    int   addr_modes_mask, reg;
    long  size, src, dst, res;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : ALL_ADDR;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, addr_modes_mask))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        src = D[reg] & size;
        get(EA1, &dst, size);
        res = dst - src;
        put(EA1, res, size);
    }
    else  {
        get(EA1, &src, size);
        dst = D[reg] & size;
        res = dst - src;
        put(&D[reg], res, size);
    }
    ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    TrcPrintf((",D%d", reg));

    return SUCCESS;
}


int SUBA(void)
{
    int   reg;
    long  size, src, dst;

    size = (inst & 0x0100) ? SZ_LONG : SZ_WORD;
	TrcPrintf((".%c\t", (size == SZ_LONG) ? 'l' : 'w'));

    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;
    get(EA1, &src, size);

    dst = A[aReg(reg)];

    put(&A[aReg(reg)], dst - src, size);

    TrcPrintf((",A%d", reg));

    return SUCCESS;
}


int SUBI(void)
{
    long  size, src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;

    memRequest(&PC, size, &src);

    TrcPrintf(("#%lx,", src));

    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;
    get(EA1, &dst, size);

    res = dst - src;
    put(EA1, res, size);
    ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    return SUCCESS;
}


int SUBQ(void)
{
    long  size, src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;

    if ((inst & 0x0038) == 0x0008)
        size = SZ_LONG;

    src = ((inst >> 9) & 0x07);
    if (src == 0)  src = 8;

    TrcPrintf(("#%ld,", src));

    if (effAddr(size, ALTERABLE_ADDR))  return ILLEGAL_INST;
    get(EA1, &dst, size);

    res = dst - src;
    put(EA1, res, size);

    if (!((inst & 0x0038) == 0x0008))
        ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    return SUCCESS;
}


int SUBX(void)
{
    long  size, src, dst, res;
    int   Rx, Ry;

    if (decodeSize(&size))  return ILLEGAL_INST;

    Ry = (inst >> 9) & 0x0007;
    Rx = inst & 0x0007;

    if (inst & 0x0008)  {
        Rx = aReg(Rx);
        Ry = aReg(Ry);
        if (size == SZ_LONG)  {
            A[Rx] -= 4;
            A[Ry] -= 4;
        }
        else if (size == SZ_WORD)  {
            A[Rx] -= 2;
            A[Ry] -= 2;
        }
        else  {
            A[Rx] -= 1;
            A[Ry] -= 1;
        }
        memGet(A[Rx], size, &src);
        memGet(A[Ry], size, &dst);
        put((long *)A[Ry], dst - src - ((SR & xbit)>> 4), size);
        memGet(A[Ry], size, &res);
    }
    else  {
        src = D[Rx] & size;
        dst = D[Ry] & size;
        put(&D[Ry], dst - src - ((SR & xbit) >> 4), size);
        res = D[Ry] & size;
    }

    ccrUpdate(GEN, GEN, CASE_1, CASE_2, CASE_6, src, dst, res, size, 0);

    return SUCCESS;
}


int DIVS(void)
{
    int   reg;
    long  src, dst, res;
    int   overflow;
    long  remainder;

    reg = (inst >> 9) & 0x0007;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;
    get(EA1, &src, SZ_WORD);

    signExtend(src, SZ_WORD, &src);
    signExtend(D[reg], SZ_LONG, &dst);

    if (src == 0)  return DIV_BY_ZERO;

    /* Check for overflow */
    if (((src < 0) && (dst >= 0)) || ((src >= 0) && (dst < 0))) {
        if ((dst / -src) > 0x0000FFFF)
            overflow = 1;
        else
            overflow = 0;
    }
    else  {
        if ((dst / src) > 0x0000FFFF)
            overflow = 1;
        else
            overflow = 0;
    }

    if (overflow)
        SR |= vbit;
    else  {
        SR &= ~vbit;
        res = (dst / src) & 0x0000FFFF;
        remainder = (dst % src) & 0x0000FFFF;
        D[reg] = res = res | (remainder * 0x00010000);
        ccrUpdate(N_A, GEN, GEN, N_A, ZER, src, dst, res, SZ_WORD, 0);
        if (((src < 0) && (dst >= 0)) || ((src >= 0) && (dst < 0)))
            SR |= nbit;
        else
            SR &= ~nbit;
    }

    return SUCCESS;
}


int DIVU(void)
{
    int   reg;
    long  src, dst, res;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;
    reg = (inst >> 9) & 0x0007;
    get(EA1, &src, SZ_WORD);
    dst = D[reg];

    if (src == 0)  return DIV_BY_ZERO;

    if ((dst / src) > 0x0000FFFF)  {
        SR |= vbit;
    }
    else  {
        SR &= ~vbit;
        D[reg] = res = ((dst / src) & 0x0000ffff) | ((dst % src) * 0x00010000);
        ccrUpdate(N_A, GEN, GEN, N_A, ZER, src, dst, res, SZ_WORD, 0);
    }

    return SUCCESS;
}


int MULS(void)
{
    int   reg;
    long  src, dst, res;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;
    reg = (inst >> 9) & 0x0007;
    get(EA1, &src, SZ_WORD);
    signExtend(src, SZ_WORD, &src);
    signExtend(D[reg], SZ_WORD, &dst);

    D[reg] = res = src * dst;

    ccrUpdate(N_A, GEN, GEN, CASE_9, ZER, src, dst, res, SZ_WORD, 0);

    return SUCCESS;
}


int MULU(void)
{
    int   reg;
    long  src, dst, res;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;
    reg = (inst >> 9) & 0x0007;
    get(EA1, &src, SZ_WORD);
    dst = D[reg] & 0x0000ffff;

    D[reg] = res = src * dst;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, SZ_LONG, 0);

    return SUCCESS;
}


int NEG(void)
{
    long  size;
    long src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr (size, DATA_ALT_ADDR))  return ILLEGAL_INST;
    get(EA1, &src, size);

    res = dst = -src;
    put(EA1, res, size);

    ccrUpdate(GEN, GEN, GEN, CASE_3, CASE_4, src, dst, res, size, 0);

    return SUCCESS;
}


int NEGX(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;
    get(EA1, &src, size);

    res = dst = -src - ((SR &xbit) >> 4);
    put(EA1, res, size);

    ccrUpdate(GEN, GEN, CASE_1, CASE_3, CASE_4, src, dst, res, size, 0);

    return SUCCESS;
}


int CMP(void)
{
    int   reg;
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;
    get(EA1, &src, size);
    dst = D[reg] & size;
    res = dst - src;
    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    TrcPrintf((",D%d", reg));

    return SUCCESS;
}


int CMPA(void)
{
    int   reg;
    long  size;
    long  src, dst, res;

    size = (inst & 0x0100) ? SZ_LONG : SZ_WORD;
    TrcPrintf((".%c\t", (size == SZ_LONG) ? 'l' : 'w'));

    if (effAddr(size, ALL_ADDR))  return ILLEGAL_INST;

    reg = aReg((inst >> 9) & 0x0007);
    get(EA1, &src, size);
    dst = A[reg] & size;
    if (size == SZ_WORD)
        signExtend(src, SZ_WORD, &src);
    res = dst - src;

    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    TrcPrintf((",A%d", reg));

    return SUCCESS;
}


int CMPI(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;

    memRequest(&PC, size, &src);
    TrcPrintf(("#%lx,", src));
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;
    get(EA1, &dst, size);
    res = dst - src;

    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    return SUCCESS;
}


int CMPM(void)
{
    long  size;
    long  src, dst, res;
    int   Rx, Ry;

    if (decodeSize(&size))  return ILLEGAL_INST;

    Rx = aReg((inst >> 9) & 0x07);
    Ry = aReg(inst & 0x07);

    memGet(A[Ry], size, &src);
    memGet(A[Rx], size, &dst);

    res = dst - src;

    if (size == SZ_BYTE)  {
        A[Rx] += 1;
        A[Ry] += 1;
    }
    else if (size == SZ_WORD)  {
        A[Rx] += 2;
        A[Ry] += 2;
    }
    else  {
        A[Rx] += 4;
        A[Ry] += 4;
    }
    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, src, dst, res, size, 0);

    return SUCCESS;
}


int TST(void)
{
    long  size;
    long  src, dst;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &src, size);

    dst = src;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, dst, size, 0);

    return SUCCESS;
}


int CLR(void)
{
    long  size;
    long  src, dst;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &src, size);

    dst = src;
    put(EA1, 0, size);
    ccrUpdate(N_A, ZER, GEN, ZER, ZER, src, dst, 0, size, 0);

    return SUCCESS;
}


int EXT(void)
{
    int   reg;
    long  size;
    long  src, dst;

    reg = inst & 0x07;

    size = (inst & 0x0040) ? SZ_LONG : SZ_WORD;
    TrcPrintf(("%s\t", (size == SZ_LONG) ? ".l" : ".w"));
    src = dst = D[reg] & size;

    if (size == SZ_WORD)
        D[reg] = (D[reg] & ~0x0000ffff) | (D[reg] & 0x000000ff) | (0x0000ff00 * ((D[reg]>>7) & 0x01));
    else
        D[reg] = (D[reg] &  0x0000ffff) | (0xffff0000 * ((D[reg] >> 15) & 0x01));

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, D[reg], size, 0);

    TrcPrintf(("D%d", reg));

    return SUCCESS;
}


int ABCD(void)
{
    long  src, dst, res;
    int   Rx, Ry, carry, temp_result;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        /* Rx & Ry are address registers used in predecrement mode */
        Rx = aReg(Rx);
        Ry = aReg(Ry);
        A[Rx]--;
        A[Ry]--;
        memGet(A[Ry], SZ_BYTE, &src);
        memGet(A[Rx], SZ_BYTE, &dst);
    }
    else {
        /* Rx & Ry are data registers */
        src = D[Ry] & 0x000000ff;
        dst = D[Rx] & 0x000000ff;
    }

    res = ((SR & xbit) >> 4) + (src & 0xf) + (dst & 0xf);
    if (res > 9)  {
        res = res - 10;
        carry = 1;
    }
    else
        carry = 0;
    temp_result = ((src >> 4) & 0xf) + ((dst >> 4) & 0xf) + carry;
    if (temp_result > 9)  {
        temp_result = temp_result - 10;
        carry = 1;
    }
    else
        carry = 0;
    res = res + (temp_result << 4);

    if (inst & 0x0008)
        memPut(A[Rx], res, SZ_BYTE);
    else
        put(&D[Rx], res, SZ_BYTE);
    if (carry)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    ccrUpdate(GEN, UND, CASE_1, UND, N_A, src, dst, res, (long) SZ_BYTE, 0);

    return SUCCESS;
}


int SBCD(void)
{
    long  src, dst, res;
    int   Rx, Ry, borrow, temp_result;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        /* Rx & Ry are address registers used in predecrement mode */
        Rx = aReg(Rx);
        Ry = aReg(Ry);
        A[Rx]--;
        A[Ry]--;
        memGet(A[Ry], SZ_BYTE, &src);
        memGet(A[Rx], SZ_BYTE, &dst);
    }
    else {
        /* Rx & Ry are data registers */
        src = D[Ry];
        dst = D[Rx];
    }

    res = (dst & 0xf) - (src & 0xf) - ((SR & xbit) >> 4);
    if (res < 0)  {
        res = res + 10;
        borrow = 1;
    }
    else
        borrow = 0;
    temp_result = ((dst >> 4) & 0xf) - ((src >> 4) & 0xf) - borrow;
    if (temp_result < 0)  {
        temp_result = temp_result + 10;
        borrow = 1;
    }
    else
        borrow = 0;
    res = res + (temp_result << 4);

    if (inst & 0x0008)
        memPut(A[Rx], res, SZ_BYTE);
    else
        put(&D[Rx], res, SZ_BYTE);

    if (borrow)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    ccrUpdate(GEN, UND, CASE_1, UND, N_A, src, dst, res, SZ_BYTE, 0);

    return SUCCESS;
}


int NBCD(void)
{
    long  src, res;
    int   borrow, temp_result;

    if (effAddr ((long) SZ_BYTE, DATA_ALT_ADDR))  return ILLEGAL_INST;
    get(EA1, &src, SZ_BYTE);

    res = 10 - (src & 0xf) - ((SR & xbit) >> 4);
    borrow = (res < 10) ? 1 : 0;
    temp_result = 10 - ((src >> 4) & 0xf) - borrow;
    borrow = (temp_result < 10) ? 1 : 0;
    res = res + (temp_result << 4);
    if (borrow)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    put(EA1, res, SZ_BYTE);
    ccrUpdate(GEN, UND, CASE_1, UND, N_A, src, res, res, SZ_BYTE, 0);

    return SUCCESS;
}


int AND(void)
{
    int   reg;
    long  size;
    long  src, dst, res;
    int   addr_modes_mask;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : DATA_ADDR;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, addr_modes_mask))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        src = D[reg] & size;
        get(EA1, &dst, size);
        res = dst & src;
        put(EA1, res, size);
    }
    else {
        get(EA1, &src, size);
        dst = D[reg] & size;
        res = dst & src;
        put(&D[reg], res, size);
    }
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int ANDI(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;

    memRequest(&PC, size, &src);

    TrcPrintf(("#%lx,", src));

    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    res = src & dst;
    put(EA1, res, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int ANDI_CCR(void)
{
    long  temp;

    memRequest(&PC, SZ_WORD, &temp);

    SR &= (temp | 0xff00);

    TrcPrintf(("#%02x,CCR", (unsigned char) temp));

    return SUCCESS;
}


int ANDI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, SZ_WORD, &temp);
    SR &= temp;

    TrcPrintf(("#%04x,", (unsigned short) temp));

    return SUCCESS_PRIV;
}


int OR(void)
{
    int   reg;
    long  size;
    long  src, dst, res;
    int   addr_mode_mask;

    addr_mode_mask = (inst & 0x0100) ? MEM_ALT_ADDR : DATA_ADDR;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, addr_mode_mask))  return ILLEGAL_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        src = D[reg] & size;
        get(EA1, &dst, size);
        res = src | dst;
        put(EA1, res, size);
    }
    else {
        get(EA1, &src, size);
        dst = D[reg] & size;
        res = src | dst;
        put(&D[reg], res, size);
    }
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int ORI(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    memRequest(&PC, size, &src);

    TrcPrintf(("#%lx,", src));

    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    res = src | dst;
    put(EA1, res, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int ORI_CCR(void)
{
    long  temp;

    memRequest(&PC, SZ_WORD, &temp);
    SR |= (temp & 0x000000ff);

    TrcPrintf(("%02x,CCR", (unsigned char) temp));

    return SUCCESS;
}


int ORI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, SZ_WORD, &temp);
    SR |= temp;

    TrcPrintf(("%04x,SR", (unsigned short) temp));

    return SUCCESS_PRIV;
}


int EOR(void)
{
    int   reg;
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;
    reg = (inst >> 9) & 0x0007;

    src = D[reg] & size;
    get(EA1, &dst, size);
    res = src ^ dst;
    put(EA1, res, size);

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int EORI(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    memRequest(&PC, size, &src);

    TrcPrintf(("#%lx,", src));

    if (effAddr(size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    res = src ^ dst;
    put(EA1, res, size);

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}


int EORI_CCR(void)
{
    long  temp;

    memRequest(&PC, SZ_WORD, &temp);

    SR ^= (temp & 0x0000ffff);

    TrcPrintf(("%02x,CCR", (unsigned char) temp));

    return SUCCESS;
}


int EORI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, SZ_WORD, &temp);
    SR ^= temp;

    TrcPrintf(("%04x,SR", (unsigned short) temp));

    return SUCCESS_PRIV;
}


int NOT(void)
{
    long  size;
    long  src, dst, res;

    if (decodeSize(&size))  return ILLEGAL_INST;
    if (effAddr (size, DATA_ALT_ADDR))  return ILLEGAL_INST;

    get(EA1, &dst, size);
    src = dst;
    res = ~dst;
    put(EA1, res, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, size, 0);

    return SUCCESS;
}



int ASR(void)
{
    long    size;
    long    src, dst;
    int     shiftCnt, cnt, reg;
    u_long  msbMask;
    u_long  tempBit;
    u_long  ev1;


    if (decodeSize(&size))  {
        /* Memory shift */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
     }
    else  {
        /* Register shift */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    msbMask = SZ_MSB(size);

    for (cnt = 1; cnt <= shiftCnt; cnt++)  {
        tempBit = ev1 & msbMask;
        ev1 = (ev1 >> 1);
        if (tempBit)  ev1 |= msbMask; 
    }
    put(EA1, ev1, size);
    ccrUpdate (GEN, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);

    return SUCCESS;
}


int ASL(void)
{
    long  size;
    long  src, dst;
    int   shiftCnt, reg;
    long  ev1;


    if (decodeSize(&size))  {
        /* Memory shift */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
     }
    else  {
        /* Register shift */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    put(EA1, (ev1 & size) << shiftCnt, size);
    get(EA1, &ev1, size);
    ccrUpdate(GEN, GEN, GEN, CASE_4, CASE_3, src, dst, ev1, size, shiftCnt);

    return SUCCESS;
}


int LSR(void)
{
    long  size;
    long  src, dst;
    int   shiftCnt, reg;
    long  ev1;


    if (decodeSize(&size))  {
        /* Memory shift */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
     }
    else  {
        /* Register shift */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    put(EA1, ((unsigned long)ev1 & size) >> shiftCnt, size);
    get(EA1, &ev1, size);
    ccrUpdate (GEN, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);

    return SUCCESS;
}



int LSL(void)
{
    long  size;
    long  src, dst;
    int   shiftCnt, reg;
    long  ev1;


    if (decodeSize(&size))  {
        /* Memory shift */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
     }
    else  {
        /* Register shift */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    put(EA1, (unsigned long)ev1 << shiftCnt, size);
    get(EA1, &ev1, size);
    ccrUpdate(GEN, GEN, GEN, ZER, CASE_3, src, dst, ev1, size, shiftCnt);

    return SUCCESS;
}


int ROXR(void)
{
    long  size;
    long  src, dst;
    int   shiftCnt, reg, cnt;
    u_long  msbMask;
    u_long  ev1;
    u_long  tempBit, tempXBit;


    if (decodeSize(&size))  {
        /* Memory rotate */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return  ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
    }
    else  {
        /* Register rotate */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    msbMask = SZ_MSB(size);

    for (cnt = 1; cnt <= shiftCnt; cnt++)  {
        tempBit = ev1 & 1;
        tempXBit = (SR & xbit) >> 4;
        ev1 = (ev1 >> 1);
        if (tempXBit)  ev1 |= msbMask;
        SR = (tempBit) ? (SR | xbit) : (SR & ~xbit);
    }
    put(EA1, ev1, size);
    ccrUpdate(GEN, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);

    return SUCCESS;
}



int ROXL(void)
{
    long  size;
    long  src, dst;
    int   shiftCnt, reg, cnt;
    u_long  msbMask;
    u_long  ev1;
    u_long  tempBit, tempXBit;


    if (decodeSize(&size))  {
        /* Memory rotate */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return  ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
    }
    else  {
        /* Register rotate */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    msbMask = SZ_MSB(size);

    for (cnt = 1; cnt <= shiftCnt; cnt++)  {
        tempBit = ev1 & msbMask;
        tempXBit = (SR & xbit) >> 4;
        ev1 = ((ev1 << 1) & 0xfffffffe);
        if (tempXBit)  ev1 |= 0x00000001;
        SR = (tempBit) ? (SR | xbit) : (SR & ~xbit);
    }
    put(EA1, ev1, size);
    ccrUpdate(N_A, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);
 
    return SUCCESS;
}


int ROR(void)
{
    long    size;
    long    src, dst;
    int     shiftCnt, reg, cnt;
    u_long  msbMask;
    u_long  ev1;
    u_long  tempBit;


    if (decodeSize(&size))  {
        /* Memory rotate */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return  ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
    }
    else  {
        /* Register rotate */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    msbMask = SZ_MSB(size);

    for (cnt = 1; cnt <= shiftCnt; cnt++)  {
        tempBit = ev1 & 1;
        ev1 = (ev1 >> 1);
        if (tempBit)  ev1 |= msbMask;
    }
    put(EA1, ev1, size);
    ccrUpdate(N_A, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);
 
    return SUCCESS;
}



int ROL(void)
{
    long    size;
    long    src, dst;
    int     shiftCnt, reg, cnt;
    u_long  msbMask;
    u_long  ev1;
    u_long  tempBit;


    if (decodeSize(&size))  {
        /* Memory rotate */
        if (effAddr(SZ_WORD, MEM_ALT_ADDR))  return  ILLEGAL_INST;
        size = SZ_WORD;
        shiftCnt = 1;
        get(EA1, &ev1, size);
        src = dst = ev1;
    }
    else  {
        /* Register rotate */
        if (inst & 0x0020) {
            shiftCnt = D[(inst >> 9) & 0x7] % 64;
            TrcPrintf(("D%d,", (int)(inst & 0x0e00) >> 9));
        }
        else {
            shiftCnt = (inst >> 9) & 0x7;
            if (shiftCnt == 0)  shiftCnt = 8;
            TrcPrintf(("#%d,", shiftCnt));
        }
        reg = inst & 0x0007;
        TrcPrintf(("D%d", reg));
        src = dst = D[reg] & size;
        EA1 = &D[reg];
        get(EA1, &ev1, size);
    }

    msbMask = SZ_MSB(size);

    for (cnt = 1; cnt <= shiftCnt; cnt++)  {
        tempBit = ev1 & msbMask;
        ev1 = ((ev1 << 1) & 0xfffffffe);
        if (tempBit)  ev1 |= 0x00000001;
    }
    put(EA1, ev1, size);
    ccrUpdate(N_A, GEN, GEN, ZER, CASE_2, src, dst, ev1, size, shiftCnt);
 
    return SUCCESS;
}



int SWAP(void)
{
    int   reg;
    long  src, dst, res;

    reg = inst & 0x07;

    src = dst = D[reg];
    res = ((src & 0x0000ffff) << 16) | ((src & 0xffff0000) >> 16);
    D[reg] = res;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, SZ_LONG, 0);

    TrcPrintf(("D%d", reg));

    return SUCCESS;
}


int BITOP(void)
{
    long  size, bit;
    long  dst;

    if (inst & 0x0100) {
        bit = D[(inst >> 9) & 0x07];
        TrcPrintf(("D%ld,", (inst >> 9) & 0x07));
    }
    else {
        memRequest(&PC, SZ_WORD, &bit);
        bit = bit & 0xff;
        TrcPrintf(("#%ld,", bit));
    }

    if (effAddr(SZ_BYTE, DATA_ADDR))  return  ILLEGAL_INST;

    if (inst & 0x0038)  {
       bit &= 0x0007;
       size = SZ_BYTE;
    }
    else {
       bit &= 0x001F;
       size = SZ_LONG;
    }

    get(EA1, &dst, size);
    if (dst & (1 << bit))
        SR &= ~zbit;
    else
        SR |= zbit;

    switch ((inst >> 6) & 0x3) {
      case 0 :                  /* perform a bit test operation */
        break;

      case 1 :                  /* perform a bit change operation */
        if (dst & (1 << bit))
            put(EA1, dst & (~(1 << bit)), size);
        else
            put(EA1, dst | (1 << bit), size);
        break;

      case 2 :                  /* perform a bit clear operation */
        put(EA1, dst & (~(1 << bit)), size);
        break;

      case 3 :                  /* perform a bit set operation */
        put(EA1, dst | (1 << bit), size);
        break;
    }

    return SUCCESS;
}


int TAS(void)
{
    long   src, dst, res;

    if (effAddr(SZ_BYTE, DATA_ALT_ADDR))  return  ILLEGAL_INST;

    get(EA1, &src, SZ_BYTE);
    res = dst = src | 0x80;
    put(EA1, res, SZ_BYTE);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, res, SZ_BYTE, 0);

    return SUCCESS;
}


int BCC(void)
{
    long  disp;
    int   branch;

    branch = checkCondition((inst & 0x0f00) >> 8);

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, SZ_WORD, &disp);
        signExtend(disp, SZ_WORD, &disp);
        if (branch)
            PC += disp;
        else
            PC += 2;
        TrcPrintf((".w\t%04x", (unsigned short) disp));
    }
    else  {
        signExtend(disp, SZ_BYTE, &disp);
        if (branch)
            PC += disp;
        TrcPrintf((".s\t%02x", (unsigned short) disp));
    }

    return SUCCESS;
}


int DBCC(void)
{
     long  disp;
     int   reg;
     int   oprt;

     memRequest(&PC, SZ_WORD, &disp);
     signExtend(disp, SZ_WORD, &disp);
     reg = inst & 0x0007;

     if (!checkCondition((inst >> 8) & 0x0f))  {
         oprt = (D[reg] & 0x0000ffff);
         oprt--;
         put(&D[reg], oprt, SZ_WORD);
         if (oprt != -1) {
             PC += (disp - 2);
         }
    }
    TrcPrintf(("\tD%d,%04x", reg, (unsigned short) disp));

    return SUCCESS;
}


int SCC(void)
{
    if (effAddr(SZ_BYTE, DATA_ALT_ADDR))
        return  ILLEGAL_INST;

    if (checkCondition((inst & 0x0f00) >> 8))
        put(EA1, 0xff, SZ_BYTE);
    else
        put(EA1, 0x00, SZ_BYTE);

    return SUCCESS;
}


int BRA(void)
{
    long  disp;

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, SZ_WORD, &disp);
        signExtend(disp, SZ_WORD, &disp);
        PC += disp;
        TrcPrintf((".w   %04x", (unsigned short) disp));
    }
    else  {
        signExtend(disp, SZ_BYTE, &disp);
        PC += disp;
        TrcPrintf((".s   %02x", (unsigned char) disp));
    }

    return SUCCESS;
}


int BSR(void)
{
    long  disp;

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, SZ_WORD, &disp);
        signExtend(disp, SZ_WORD, &disp);
        A[aReg(7)] -= 4;
        memPut(PC+2, A[aReg(7)], SZ_LONG);
        PC += disp;
        TrcPrintf((".w   %04x", (unsigned short) disp));
    }
    else  {
        signExtend(disp, SZ_BYTE, &disp);
        A[aReg(7)] -= 4;
        memPut(PC, A[aReg(7)], SZ_LONG);
        PC += disp;
        TrcPrintf((".s   %02x", (unsigned char) disp));
    }

    return SUCCESS;
}


int JMP(void)
{
    if (effAddr(SZ_WORD, CONTROL_ADDR))
        return  ILLEGAL_INST;

    PC = *EA1;

    return SUCCESS;
}


int JSR(void)
{
    if (effAddr(SZ_WORD, CONTROL_ADDR))
        return  ILLEGAL_INST;

    A[aReg(7)] -= 4; memPut(PC, A[aReg(7)], SZ_LONG);
    PC = *EA1;

    return SUCCESS;
}


int RTE(void)
{
    long  temp;

    if (!(SR & sbit))
        return (NO_PRIVILEGE);

    memGet(A[8], SZ_WORD, &temp);
    A[8] += 2;
    SR = temp & 0x0000ffff;
    memGet(A[8], SZ_LONG, &PC);
    A[8] += 4;

    return SUCCESS_PRIV;
}


int RTR(void)
{
    long  temp;

    memGet(A[aReg(7)], SZ_WORD, &temp);
    A[aReg(7)] += 2;
    SR = (SR & 0xff00) | (temp & 0x00ff);
    memGet(A[aReg(7)], SZ_LONG, &PC);
    A[aReg(7)] += 4;

    return SUCCESS_PRIV;
}


int RTS(void)
{
    memGet(A[aReg(7)], SZ_LONG, &PC);
    A[aReg(7)] += 4;

    return SUCCESS;
}


int NOP(void)
{
    return SUCCESS;
}


int CHK(void)
{
    int   reg;
    long  src, dst;

    if (effAddr(SZ_WORD, DATA_ADDR))  return ILLEGAL_INST;
    reg = (inst >> 9) & 0x07;

    get(EA1, &src, SZ_WORD);
    signExtend(src, SZ_WORD, &src);
    signExtend(D[reg], SZ_WORD, &dst);

    ccrUpdate(N_A, GEN, UND, UND, UND, src, dst, dst, SZ_WORD, 0);

    if ((dst < 0) || (dst > src))
        return CHK_EXCEPTION;

    return SUCCESS;
}


int ILLEGAL(void)
{
    return ILLEGAL_INST;
}


int RESET(void)
{
    if (!(SR & sbit))
        return NO_PRIVILEGE;

    return SUCCESS;
}



int STOP(void)
{
    long  temp;
    int   tr_on;

    memRequest(&PC, SZ_WORD, &temp);

    tr_on = (SR & tbit) ? 1 : 0;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    SR = temp & 0x0000ffff;
    if (tr_on)
        SR = SR | tbit;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    cpuState = ST_STOP;

    return SUCCESS_PRIV;
}



int TRAP(void)
{
    TrcPrintf(("#%d\n", (int)(inst & 0x000F)));

    return TRAP_TRAP;
}



int TRAPV(void)
{
    if (SR & vbit)
        return TRAPV_TRAP;

     return SUCCESS;
}


int LINE1010(void)
{
    PC -= 2;
    return LINE_1010;
}


int LINE1111(void)
{
    PC -= 2;
    return LINE_1111;
}


#include "opcodetab.h"

