/*
 *  Copyright (C) 1997-2001 B. Cameron Lesiuk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ECMA-48 Implementation Notes
 * - Modes: EMCA-48 depreciates the use of modes. Consequently, the use of
 *   modes are not implemented. All modes are assumed to have their default value.
 * - It is unspecified what should happen when a LF is given at the bottom of
 *   the screen (or I'm missing something). Anyways, I coded it so that it scrolls
 *   the screen; I think everyone and their dog expects this anyways.
 * - FF is technically unimplemented; it does the same thing as an LF.
 * - Only align-left tabs are implemented. Affected operations: HT TAC TALE TATE TCC
 *   (and probably a few others)
 * - SIMD is implemented, but I'm not sure if I did it completely or correctly.
 *   Setting SIMD to 1 affects *ONLY* BS, CR, NEL, and the printing of characters.
 *   With SIMD=1, characters will be printed in a right-to-left progression, complete
 *   with wrap up and line-feeds when the cursor gets to the top of the screen.
 *   I am unsure if this is what ECMA-48 intended. Furthermore, I am unsure what
 *   other operations, if any, should be affected by SIMD.
 * - I'm not sure if CNL is supposed to generate linefeeds. Currently, it does.
 *   Same goes for CPL.
 * - I didn't have a *clue* what to do with F1-F12, INS, DEL, Home, End, PgUp,
 *   PgDn, or the cursor keys. So I guessed:
 *      INS  = ESC [ 2 ~         F1 = ESC [ [ A        F11= ESC [ 23 ~
 *      DEL  = DEL (0x7F)        F2 = ESC [ [ B        F12= ESC [ 24 ~
 *      Home = ESC [ H           F3 = ESC [ [ C
 *      End  = ESC [ 4 ~         F4 = ESC [ [ D
 *      PgUp = ESC [ 5 ~         F5 = ESC [ [ E
 *      PgDn = ESC [ 6 ~         F6 = ESC [ 17 ~
 *      Up   = ESC [ A           F7 = ESC [ 18 ~
 *      Down = ESC [ B           F8 = ESC [ 19 ~
 *      Right= ESC [ C           F9 = ESC [ 20 ~
 *      Left = ESC [ D           F10= ESC [ 21 ~
 *
 */


#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <stdio.h>

#include "resource.h"
#include "termbuf.h"
#include "ansiterm.h"
#include "ansitbl.h"
#include "ascii.h"
#include "enviromt.h"

#define AT_DEFAULT_TITLE  "vt100"
#define AT_DEFAULT_X      80
#define AT_DEFAULT_Y      25
#define AT_MIN_X          20       /* Minimum X window size (in characters) */
#define AT_MIN_Y          1        /* Minimum Y window size (in characters) */

/* The TermBuf doesn't work with a scrollback of 0, so this helps it along some */
#define ATIB_TERMBUF_EXTRALINES 5    /* this is not necessary if termbuf handled scrollback=0 */

/* Timer IDs */
#define AT_TIMER_SCROLL 1  /* controlls speed of scroll when mouse is outside the window */
#define AT_TIMER_FLASH  2  /* controlls the speed of a visual bell (window flash) */




void vt100DestroyWnd(ATTERM *atTerm)
{
	HGLOBAL memory;

	if (!atTerm)  return;

	DestroyWindow(atTerm->aWnd);
	tbFreeTermBuf(atTerm->aTermBuf);

	memory = atTerm->aOutBufMemory;
	GlobalUnlock(memory);
	GlobalFree(memory);
	memory = atTerm->aMemory;
	GlobalUnlock(memory);
	GlobalFree(memory);
}


void atSetFont(ATTERM *atTerm, HFONT font, LOGFONT *logFont)
{
	HDC        hDC;
	TEXTMETRIC tm;
	HFONT      newfont;

	if (!atTerm)  return;

	if (atTerm->aFont) {
		DeleteObject(atTerm->aFont);
	}
	if (atTerm->aFontUnderline) {
		DeleteObject(atTerm->aFontUnderline);
	}
	if (logFont) {
		newfont = CreateFontIndirect(logFont);
		if (newfont) {
			atTerm->aFont = newfont;
			memcpy(&(atTerm->aLogFont), logFont, sizeof(LOGFONT));
			atTerm->aLogFontValid = TRUE;
			if (logFont->lfUnderline) {
				atTerm->aFontUnderline = NULL;
			} else {
				logFont->lfUnderline = TRUE;
				newfont = CreateFontIndirect(logFont);
				atTerm->aFontUnderline = newfont;
			}
		} else {
			atTerm->aFont = GetStockObject(SYSTEM_FIXED_FONT);
			atTerm->aLogFontValid  = FALSE;
			atTerm->aFontUnderline = NULL;
		}
	} else if (font) {
		atTerm->aFont = font;
		atTerm->aLogFontValid  = FALSE;
		atTerm->aFontUnderline = NULL;
	} else {
		atTerm->aFont = GetStockObject(SYSTEM_FIXED_FONT);
		atTerm->aLogFontValid  = FALSE;
		atTerm->aFontUnderline = NULL;
	}

	if (atTerm->aFont) {
		hDC = GetDC(atTerm->aWnd);
		if (hDC) {
			SelectObject(hDC, atTerm->aFont);
			GetTextMetrics(hDC, &tm);
			ReleaseDC(atTerm->aWnd, hDC);
			atTerm->aFontX = tm.tmAveCharWidth;
			atTerm->aFontY = tm.tmHeight;
		}
	} else {
		atTerm->aFontX = atTerm->aFontY = 1;
	}

}

	
void vt100ChangeFont(ATTERM *atTerm, HFONT font, LOGFONT *logFont)
{
	RECT   rect;
    int    xe, ye;
	if (!atTerm)  return;

	atSetFont(atTerm, font, logFont);

	GetWindowRect(atTerm->aWnd, &rect);
    xe = 2 * GetSystemMetrics(SM_CXFRAME) + GetSystemMetrics(SM_CXVSCROLL) + GetSystemMetrics(SM_CXBORDER);
    ye = 2 * GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION);
	MoveWindow(atTerm->aWnd, rect.left, rect.top,
		atTerm->aSizeX * atTerm->aFontX + xe,
		atTerm->aSizeY * atTerm->aFontY + ye, TRUE);
	InvalidateRect(atTerm->aWnd, NULL, TRUE);
	if (!atTerm->aFontX)  atTerm->aFontX = 2;
	if (!atTerm->aFontY)  atTerm->aFontY = 2;
	/* Resize caret */
	DestroyCaret();
	CreateCaret(atTerm->aWnd, NULL, atTerm->aFontX, atTerm->aFontY);
	atUpdateCaret(atTerm);
	/* Resize underline pen */
	tbCreateUnderscorePen(atTerm->aTermBuf, atTerm->aFontY);
}


void atUpdateTermTitle(ATTERM *atTerm, char *newTitle)
{
	char buf[TBSTR_TITLE_LEN+50];

	if (!atTerm)  return;

	if (newTitle)
			strcpy(atTerm->aTitle, newTitle);
	sprintf(buf, "%s (%ix%i)", atTerm->aTitle, atTerm->aSizeX, atTerm->aSizeY);
	SetWindowText(atTerm->aWnd, buf);
}


void atAdjustScrollbar(ATTERM *atTerm)
{
	if (atTerm) {
		SetScrollRange(atTerm->aWnd, SB_VERT,
			atTerm->aSizeY-atTerm->aTermBuf->tBufY + ATIB_TERMBUF_EXTRALINES, 0, FALSE);
		SetScrollPos(atTerm->aWnd, SB_VERT, atTerm->aShowY, TRUE);
	}
}


/*
 * This procedure changes all the internal ansiterm states to reflect the
 * current window dimentions
 */
void atAdjustTerm(ATTERM *atTerm, long scrollBack)
{
	RECT  rect;
	int   scrX, scrY;
	long  BufY;

	if (!atTerm)  return;

	GetClientRect(atTerm->aWnd, &rect);
	scrX = (rect.right  - rect.left) / atTerm->aFontX;
	scrY = (rect.bottom - rect.top) / atTerm->aFontY;
	if (scrX < AT_MIN_X)  scrX = AT_MIN_X;
	if (scrY < AT_MIN_Y)  scrY = AT_MIN_Y;
	if (scrollBack >= 0)
		BufY = scrY + scrollBack;
	else
		BufY = scrY + (atTerm->aTermBuf->tBufY - ATIB_TERMBUF_EXTRALINES - atTerm->aTermBuf->tScrY);
	atTerm->aTermBuf = tbReAllocTermBuf(atTerm->aTermBuf, scrX,
							BufY + ATIB_TERMBUF_EXTRALINES, scrX, scrY);
	atTerm->aSizeX = atTerm->aTermBuf->tScrX;
	atTerm->aSizeY = atTerm->aTermBuf->tScrY;

	/* Reset scrolling */
	atTerm->aShowX = atTerm->aShowY = 0;

	/* Reposition scrollbars accordingly. */
	atAdjustScrollbar(atTerm);

	/* Redraw the entire box */
	InvalidateRect(atTerm->aWnd, NULL, TRUE);
	atUpdateTermTitle(atTerm, NULL);

	/* Notify our owner of any changes */
	if (atTerm->aNotify)  (*(atTerm->aNotify))(atTerm, ATNOTIFY_WINDOWSIZE);
}


void atShowTerm(ATTERM *atTerm)
{
	if (atTerm)
		ShowWindow(atTerm->aWnd, SW_SHOWNORMAL);
}


void atHideTerm(ATTERM *atTerm)
{
	if (atTerm)
		ShowWindow(atTerm->aWnd, SW_HIDE);
}


BOOL atIsTermVisible(ATTERM *atTerm)
{
	if (atTerm)
		return IsWindowVisible(atTerm->aWnd);
	return FALSE;
}


int vt100Send(ATTERM *atTerm, unsigned char *pBuf, int pBufLen)
{
	int   byte, i, j;
	int   result;
	long  x, y;
	unsigned long p1, p2, p3;

 	if (!atTerm)  return 0;

	if (atTerm->aEmulation == AT_EMU_NONE) {
		for (byte = 0; byte < pBufLen; byte++) {
			if ((pBuf[byte] < 0x20) || (pBuf[byte] >= 0x80)) {
				switch (pBuf[byte]) {
					case 0x0A: // linefeed
						tbLineFeed(atTerm->aTermBuf);
						break;
					case 0x0D: // carridge return
//						tbCarridgeReturn(atTerm->aTermBuf);
						break;
					default:
						break;
				}
			} else {
				tbPutChar(atTerm->aTermBuf, pBuf[byte]);
			}
		}
	} else if (atTerm->aEmulation == AT_EMU_VT100) {
		for (byte = 0; byte < pBufLen; byte++) {
			if (pBuf[byte] == 0x03) {
				continue;
			}
			result = (atTerm->aState)[pBuf[byte]];
			switch (result) {
				case ATSTATE_GROUND_STATE:
					/* exit ignore mode */
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_IGNORE_STATE:
					/* IES: ignore anything else */
					atTerm->aState = atIGNTable;
					break;
				case ATSTATE_IGNORE_ESC:
					/* IGN: escape */
					atTerm->aState = atIESTable;
					break;
				case ATSTATE_IGNORE:
					/* ignore character */
					break;
				case ATSTATE_PRINT:
					/* print character */
					/* we should be using atTerm->aCharacterSet[atTerm->aCurCharacterSet] */
					if (atTerm->aSIMD)
						tbPutCharReverse(atTerm->aTermBuf, pBuf[byte]);
					else
						tbPutChar(atTerm->aTermBuf, pBuf[byte]);
					break;
				case ATSTATE_BELL:
					if (atTerm->aBell) {
						MessageBeep(MB_OK);
					} else {
						FlashWindow(atTerm->aWnd,TRUE);
						SetTimer(atTerm->aWnd, AT_TIMER_FLASH, 200, NULL);
					}
					break;
				case ATSTATE_BS:
					if (atTerm->aSIMD)
						tbCurRight(atTerm->aTermBuf);
					else
						tbCurLeft(atTerm->aTermBuf);
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DEL:
					tbDelete(atTerm->aTermBuf);
					break;
				case ATSTATE_DCH:
					/* delete character */
					i = atTerm->aStateVal[0];
					if (i < 1)
						i = 1;
					for (; i; i--) {
						tbDelete(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CR:
					if (atTerm->aSIMD)
						tbSetCursor(atTerm->aTermBuf, atTerm->aSLL, atTerm->aTermBuf->tCurY);
					else
						tbSetCursor(atTerm->aTermBuf, atTerm->aSLH, atTerm->aTermBuf->tCurY);
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_NEL:
					if (atTerm->aSIMD) {
						tbSetCursor(atTerm->aTermBuf, atTerm->aSLL, atTerm->aTermBuf->tCurY);
						tbLineFeedUp(atTerm->aTermBuf);
					}	else {
						tbSetCursor(atTerm->aTermBuf, atTerm->aSLH, atTerm->aTermBuf->tCurY);
						tbLineFeed(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_ESC:
					atTerm->aState = atESCTable;
					break;
				case ATSTATE_VT: /**/
					/* line feed, form feed, vertical tab */
					tbLineFeed(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_HT:
					/* horizontal tab */
					tbHTab(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_HTS:
					/* horizontal tab set */
					tbTabSet(atTerm->aTermBuf, TBTAB_LEFTALIGN, -1);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_TBC:
					/* horizontal tab clear */
					switch(atTerm->aStateVal[0]) {
						default:
						case 0: /* Character tab at current cursor position is cleared */
							tbTabSet(atTerm->aTermBuf, TBTAB_NONE, -1);
							break;
						case 1: /* Line tabulation at current line is cleared -- unimplemented */
						case 2: /* all character tabs in active line are cleared */
						case 3: /* all character tabs are cleared */
						case 4: /* all line tabs are cleared */
						case 5: /* all tabs are cleared (both character and line) */
							tbTabClearAll(atTerm->aTermBuf);
							break;
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_TSR:
					/* Tabulation Stop Remove */
					if (atTerm->aStateVal[0] != AT_STATEVAL_DEFAULT) {
						tbTabSet(atTerm->aTermBuf, TBTAB_NONE, atTerm->aStateVal[0]-1);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CBT:
					/* Cursor Backward Tabulation */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for(; j; j--)
						tbHTabReverse(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CHA:
					/* Cursor Character Absolute */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					tbSetCursor(atTerm->aTermBuf, j-1, atTerm->aTermBuf->tCurY);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_VPA:
					/* Line Position Absolute */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					tbSetCursor(atTerm->aTermBuf, atTerm->aTermBuf->tCurX, j-1);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CHT:
					/* Cursor Forward Tabulation */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j=1;
					for(; j; j--)
						tbHTab(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CNL:
					/* Cursor Next Line */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for(; j; j--)
						tbLineFeed(atTerm->aTermBuf);
					tbSetCursor(atTerm->aTermBuf, 0, atTerm->aTermBuf->tCurY);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CPL:
					/* Cursor Previous Line */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for(; j; j--)
						tbLineFeedUp(atTerm->aTermBuf);
					tbSetCursor(atTerm->aTermBuf, 0, atTerm->aTermBuf->tCurY);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CTC:
					/* Cursor Tabulation Control */
					/* horizontal tab clear */
					if (atTerm->aStateVal[0] == AT_STATEVAL_DEFAULT)
						atTerm->aStateVal[0] = 0;
					for (i = 0; atTerm->aStateVal[i] != AT_STATEVAL_DEFAULT; i++) {
						switch(atTerm->aStateVal[i]) {
							default:
							case 0: /* Character tab at current cursor position is set */
								tbTabSet(atTerm->aTermBuf, TBTAB_LEFTALIGN, -1);
								break;
							case 1: /* Line tabulation at current line is set -- unimplemented */
								break;
							case 2: /* Character tab at current position is cleared */
								tbTabSet(atTerm->aTermBuf, TBTAB_NONE, -1);
								break;
							case 3: /* current line tabs is cleared */
								break;
							case 4: /* all character tabs in current line are cleared */
							case 5: /* all character tabs are cleared */
								tbTabClearAll(atTerm->aTermBuf);
								break;
							case 6: /* all line tabulation stops are cleared */
								break;
						}
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SD:
					/* Scroll data down n lines */
					if (atTerm->aStateVal[0] == AT_STATEVAL_DEFAULT)
						atTerm->aStateVal[0] = 1;
					for (i = 0; i < atTerm->aStateVal[0]; i++) {
						tbScrollDown(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SU:
					/* Scroll data up n lines */
					if (atTerm->aStateVal[0] == AT_STATEVAL_DEFAULT)
						atTerm->aStateVal[0] = 1;
					for (i = 0; i < atTerm->aStateVal[0]; i++) {
						tbScrollUp(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SR:
					if (atTerm->aStateVal[0] == AT_STATEVAL_DEFAULT)
						atTerm->aStateVal[0] = 1;
					for (i = 0; i < atTerm->aStateVal[0]; i++) {
						tbScrollRight(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SL:
					if (atTerm->aStateVal[0] == AT_STATEVAL_DEFAULT)
						atTerm->aStateVal[0] = 1;
					for (i = 0; i < atTerm->aStateVal[0]; i++) {
						tbScrollLeft(atTerm->aTermBuf);
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SI:
					break;
				case ATSTATE_SO:
					break;
				case ATSTATE_SCR_STATE:
					atTerm->aState = atSCRTable;
					break;
				case ATSTATE_SCS0_STATE:
					atTerm->aSCSType = 0;
					atTerm->aState = atSCSTable;
					break;
				case ATSTATE_SCS1_STATE:
					atTerm->aSCSType = 1;
					atTerm->aState = atSCSTable;
					break;
				case ATSTATE_SCS2_STATE:
					atTerm->aSCSType = 2;
					atTerm->aState = atSCSTable;
					break;
				case ATSTATE_SCS3_STATE:
					atTerm->aSCSType = 3;
					atTerm->aState = atSCSTable;
					break;
				case ATSTATE_ESC_IGNORE:
					/* unknown ESC code */
					atTerm->aState = atEIGTable;
					break;
				case ATSTATE_ESC_DIGIT:
					j = atTerm->aStateVal[atTerm->aStateValPos];
					if (j == AT_STATEVAL_DEFAULT) {
						atTerm->aStateVal[atTerm->aStateValPos] = (pBuf[byte] - '0');
					} else {
						atTerm->aStateVal[atTerm->aStateValPos] = j*10 + (pBuf[byte] - '0');
					}
					break;
				case ATSTATE_ESC_SEMI:
					atTerm->aStateValPos++;
					break;
				case ATSTATE_ICH:
					/* insert characters */
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					tbInsertChar(atTerm->aTermBuf, j);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_RI:
					if (atTerm->aTermBuf->tCurY) {
						tbCurUp(atTerm->aTermBuf);
					} else {
						tbScrollDown(atTerm->aTermBuf);
					}
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CUU:
					j = atTerm->aStateVal[0];
					if (j < 1)
						j=1;
					for (; j; j--) {
						tbCurUp(atTerm->aTermBuf);
					}
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CUD:
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for (; j; j--) {
						tbCurDown(atTerm->aTermBuf);
					}
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CUF:
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for (; j; j--) {
						tbCurRight(atTerm->aTermBuf);
					}
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CUB:
					j = atTerm->aStateVal[0];
					if (j < 1)
						j = 1;
					for (; j; j--) {
						tbCurLeft(atTerm->aTermBuf);
					}
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CUP:
					/* cursor position */
					y = atTerm->aStateVal[0];
					x = atTerm->aStateVal[1];
					if (y < 1)  y = 1;
					if (x < 1)  x = 1;
					tbSetCursor(atTerm->aTermBuf, x-1, y-1);
					atUpdateCaret(atTerm);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_ECH:
					/* Erase Character */
					i = atTerm->aStateVal[0];
					if (i < 1)  i = 1;
					i--;
					tbClearLineX(atTerm->aTermBuf, atTerm->aTermBuf->tCurX, atTerm->aTermBuf->tCurX + i);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_ED:
					/* Erase in Page */
					switch(atTerm->aStateVal[0]) {
						default:
						case 0:
							/* clear below */
							tbClearBelow(atTerm->aTermBuf);
							break;
						case 1:
							/* clear above */
							tbClearAbove(atTerm->aTermBuf);
							break;
						case 2:
							/* clear screen */
							tbClearScreen(atTerm->aTermBuf);
							break;
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_EL:
					switch(atTerm->aStateVal[0]) {
						default:
						case 0:
							/* clear to right */
							tbClearRight(atTerm->aTermBuf);
							break;
						case 1:
							/* clear to left */
							tbClearLeft(atTerm->aTermBuf);
							break;
						case 2:
							/* clear line */
							tbClearLine(atTerm->aTermBuf);
							break;
					}
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SGR:
					if (atTerm->aStateVal[0]==AT_STATEVAL_DEFAULT) {
						/* no parameters - set defaults */
						p1 = 0x07;
						p2 = 0x00;
						p3 = 0x00;
					} else {
						p1 = TB_EXTFCOLOR(atTerm->aTermBuf->tCurrentAttribute);
						p2 = TB_EXTBCOLOR(atTerm->aTermBuf->tCurrentAttribute);
						p3 = TB_EXTEFFECT(atTerm->aTermBuf->tCurrentAttribute);
					}
					for (j = 0; atTerm->aStateVal[j] != AT_STATEVAL_DEFAULT; j++) {
						switch(atTerm->aStateVal[j]) {
							case 0:	/* default */
								/* reset - lgray on black - all attributes off */
								p1 = 0x07;
								p2 = 0x00;
								p3 = 0x00;
								break;
							case 1: /* bold on */
								p3 |= AT_FONTEFFECT_BOLD;
								break;
//							case 2:  /* faint on */
//							case 3:  /* italicized on */
							case 4:  /* singly underline on */
								/* Underscores on */
								p3 |= AT_FONTEFFECT_UNDERLINE;
								break;
							case 5:  /* slowly blinking on (less than 150 per minute) */
							case 6:  /* rapidly blinking on (more than 150 per minute) */
								/* Blink on */
//								p2 |= 0x08;  /* this should be taken out sometime */
								p3 |= AT_FONTEFFECT_BLINK;
								break;
							case 7:  /* negative image (reverse video) on */
								/* reverse video on */
								p3 |= AT_FONTEFFECT_REVERSE;
								break;
//							case 8:  /* concealed characters on */
//							case 9:  /* crossed-out (strike-out) */
							case 10: /* Primary (default) font */
							case 11: /* first alternate font */
							case 12: /* second alternate font */
							case 13: /* third alternate font */
							case 14: /* fourth alternate font */
							case 15: /* fifth alternate font */
							case 16: /* sixth alternate font */
							case 17: /* seventh alternate font */
							case 18: /* eighth alternate font */
							case 19: /* ninth alternate font */
							case 20: /* Fraktur (Gothic) */
								break;
//							case 21: /* double underline on */
							case 21: /* **>> This breaks ECMA-48 (bold off ?!?!?!) */ 
							case 22: /* normal color/intensity (neither bold nor faint) */
								/* Bold off */
//								p1 |= 0x08;
//								p1 ^= 0x08;
								p3 |= AT_FONTEFFECT_BOLD;
								p3 ^= AT_FONTEFFECT_BOLD;
								break;
//							case 23: /* not italicized, not fraktur */
							case 24: /* not underlined (not singly nor double) */
								/* Underscores off */
								p3 |= AT_FONTEFFECT_UNDERLINE;
								p3 ^= AT_FONTEFFECT_UNDERLINE;
								break;
							case 25: /* steady (blink off) */
								/* Blink off */
//								p2 |= 0x08;
//								p2 ^= 0x08;
								p3|=AT_FONTEFFECT_BLINK;
								p3^=AT_FONTEFFECT_BLINK;
								break;
//							case 26: /* (reserved for proportional spacing) */
							case 27: /* positive image (negative off) */
								/* reverse video off */
								p3 |= AT_FONTEFFECT_REVERSE;
								p3 ^= AT_FONTEFFECT_REVERSE;
								break;
//							case 28: /* revealed characters (conceal off) */
//							case 29: /* not crossed out (crossed-out off) */
							case 30: /* black display (char) */
							case 31: /* red display (char) */
							case 32: /* green display (char) */
							case 33: /* yellow display (char) */
							case 34: /* blue display (char) */
							case 35: /* magenta display (char) */
							case 36: /* cyan display (char) */
							case 37: /* white display (char) */
								/* Foreground color */
								p1 = (((p1|0x07)^0x07) | (atTerm->aStateVal[j]-30));
								break;
//							case 38: /* (reserved) */
							case 39: /* default display (char) color */
								/* turn off foreground color */
								p1 = 0x07;
								break;
							case 40: /* black background */
							case 41: /* red background */
							case 42: /* green background */
							case 43: /* yellow background */
							case 44: /* blue background */
							case 45: /* magenta background */
							case 46: /* cyan background */
							case 47: /* white background */
								/* Background color */
								p2 = (((p2|0x07)^0x07) | (atTerm->aStateVal[j]-40));
								break;
//							case 48: /* (reserved) */
							case 49: /* default background color */
								/* turn off background color */
								p2 = 0x00;
								break;
//							case 50: /* (reserved - cancels effect of 26) */
//							case 51: /* framed */
//							case 52: /* encircled */
//							case 53: /* overline */
//							case 54: /* not framed, not encircled */
//							case 55: /* not overlined */
//							case 56: /* (reserved) */
//							case 57: /* (reserved) */
//							case 58: /* (reserved) */
//							case 59: /* (reserved) */
//							case 60: /* ideogram underline or right side line */
//							case 61: /* ideogram double underline or double right side line */
//							case 62: /* ideogram overline or left side line */
//							case 63: /* ideogram double overline or double left side line */
//							case 64: /* ideogram stress marking */
//							case 65: /* cancels effect from 60-64 */
							default:
								break;
						}
					}
					tbSetFColor(atTerm->aTermBuf, p1);
					tbSetBColor(atTerm->aTermBuf, p2);
					tbSetEffect(atTerm->aTermBuf,  p3);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CSI_STATE:
					/* reset parameters */
					for (i = 0; i < AT_ESC_PARAM_MAX; i++)
						atTerm->aStateVal[i] = AT_STATEVAL_DEFAULT;
					atTerm->aStateValPos = 0;
					atTerm->aState = atCSITable;
					break;
				case ATSTATE_IL:
					j = atTerm->aStateVal[0];
					if (j < 1)  j = 1;
					tbInsertLine(atTerm->aTermBuf, j);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DL:
					j = atTerm->aStateVal[0];
					if (j < 1)  j = 1;
					tbDeleteLine(atTerm->aTermBuf, j);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DECSC:
					tbCursorSave(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DECRC:
					tbCursorRestore(atTerm->aTermBuf);
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DECSET:
				case ATSTATE_DECRST:
					/* process DEC private modes set, reset */
					if (result == ATSTATE_DECSET) {
//						bit = 0x00;  /* set bit */
					} else {
//						bit = ~0x00; /* reset (clr) bits */
					}
					i = 0;
					while(atTerm->aStateVal[i] != AT_STATEVAL_DEFAULT) {
						switch (atTerm->aStateVal[i]) {
							case 1: /* DECCKM */
							case 2: /* ANSI/VT52 mode */
							case 3: /* DECCOLM */
							case 4: /* DECSCLM (slow scroll) */
							case 5: /* DECSCNM */
							case 6: /* DECOM */
							case 7: /* DECAWM */
							case 8: /* DECARM */
							case 9: /* MIT bogus sequence */
							case 38: /* DECTEK */
							case 40: /* 132 column mode */
							case 41: /* curses hack */
							case 44: /* margin bell */
							case 45: /* reverse wraparound */
							case 46: /* logging */
							case 47: /* alternate buffer */
							case 1000: /* xterm bogus sequence */
							case 1001: /* xterm sequence w/ hilite tracking */
								break;
						}
						i++;
					}
					break;
				case ATSTATE_SIMD:
					if (atTerm->aStateVal[0] == 1)
						atTerm->aSIMD = 1;
					else
						atTerm->aSIMD = 0;
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_CSI_SIB:
					atTerm->aState = atCSI_SIBTable;
					break;
				case ATSTATE_SLH:
					if (atTerm->aStateVal[0] < 1)
						atTerm->aSLH = 0;
					else
						atTerm->aSLH = atTerm->aStateVal[0]-1;
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_SLL:
					if (atTerm->aStateVal[0] < 1)
						atTerm->aSLL = 0;
					else
						atTerm->aSLL = atTerm->aStateVal[0]-1;
					atTerm->aState = atGroundTable;
					break;
				case ATSTATE_DA1:
				case ATSTATE_TRACK_MOUSE:
				case ATSTATE_SET:
				case ATSTATE_RST:
				case ATSTATE_CPR:
				case ATSTATE_DECSTBM:
				case ATSTATE_DECREQTPARM:
				case ATSTATE_DECALN:
				case ATSTATE_GSETS:
				case ATSTATE_DECKPAM:
				case ATSTATE_DECKPNM:
				case ATSTATE_IND:
				case ATSTATE_SS2:
				case ATSTATE_SS3:
				case ATSTATE_OSC:
				case ATSTATE_RIS:
				case ATSTATE_LS2:
				case ATSTATE_LS3:
				case ATSTATE_LS3R:
				case ATSTATE_LS2R:
				case ATSTATE_LS1R:
				case ATSTATE_XTERM_SAVE:
				case ATSTATE_XTERM_RESTORE:
				case ATSTATE_XTERM_TITLE:
				case ATSTATE_DECID:
					atTerm->aState = atGroundTable;
					break;
				default:
					atTerm->aState = atGroundTable;
					break;
			} /* master switch */
		} /* for() -- loop through buffer */
	} /* VT100 emulation */

	return(pBufLen);
}


void atClearTerm(ATTERM *atTerm)
{
	if (atTerm)
		tbClearScreen(atTerm->aTermBuf);
}


void atNotifyFn(unsigned long id, unsigned int oprt, long p0, long p1, long p2, long p3)
{
	ATTERM  *term;
	long    sy, dy;
	RECT    rect;

	if (!id)  return;

	term = (ATTERM*)(id);

	switch (oprt) {
		case TBNOTIFY_SCROLL:
			term->aScroll += -p0 * term->aFontY;
			PostMessage(term->aWnd, WM_PAINT, 0, 0L); /* note there may not be an update region! He he he */
			break;

		case TBNOTIFY_NEWCHAR:
			sy = TB_MAX(p1, term->aShowY);
			dy = TB_MIN(p3, term->aShowY+term->aSizeY);
			if (dy >= sy) {
				rect.top = (sy-term->aShowY) * term->aFontY;
				rect.bottom = (sy-term->aShowY + (dy-sy) + 1) * term->aFontY;
				if (sy == dy) {
					rect.left  = (p0 - term->aShowX) * term->aFontX;
					rect.right = (p0 - term->aShowX + (p2 - p0)+1) * term->aFontX;
				} else {
					rect.left=0;
					rect.right = term->aSizeX * term->aFontX;
				}
				InvalidateRect(term->aWnd, &rect, FALSE);
			}
			break;

		case TBNOTIFY_MOVECUR:
			if (p2)
				atUpdateCaret(term);
			break;

		case TBNOTIFY_RESIZE:
            atAdjustTerm(term, -1);

		default:
			break;
	}
	return;
}


/*
 * Updates the caret position & visibility information based
 * on the current state of the terminal
 */
void atUpdateCaret(ATTERM *atTerm)
{
	if (!atTerm)  return;

	if (GetFocus() == atTerm->aWnd) {
        if (atTerm->aTermBuf) {
//			if ((atTerm->aTermBuf->tCurX>=atTerm->aShowX)&&
// 	    			(atTerm->aTermBuf->tCurX<atTerm->aShowX+atTerm->aSizeX) &&
//					(atTerm->aTermBuf->tCurY>=atTerm->aShowY)&&
//					(atTerm->aTermBuf->tCurY<atTerm->aShowY+atTerm->aSizeY)) {
				SetCaretPos((atTerm->aTermBuf->tCurX-atTerm->aShowX)*atTerm->aFontX,
						(atTerm->aTermBuf->tCurY-atTerm->aShowY)*atTerm->aFontY);
				ShowCaret(atTerm->aWnd);
//			}  else {
//				HideCaret(atTerm->aWnd);
//			}
		}
	}
}


/*
 * Does screen scrolling & invalidation etc to cause the screen to
 * be repainted at a new specified location
 */
void atScrollTermTo(ATTERM *atTerm, BOOL rel, long val)
{
	long newShowY;

	if (!atTerm)  return;

	if (rel) {
		/* relative offset */
		newShowY = atTerm->aShowY + val;
	} else {
		newShowY = val;
	}

	if (newShowY > 0)
		newShowY = 0;
	if (newShowY < (atTerm->aTermBuf->tScrY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES))
		newShowY = (atTerm->aTermBuf->tScrY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES);

	/* ok, we know where we are scrolling to. Now, let's check if we can
	 * scroll part of our window. If we can, do so. */
	if (atTerm->aShowY == newShowY) {
		return;
	} else if (atTerm->aShowY < newShowY) {
		/* scroll down */
		if ((newShowY-atTerm->aShowY) < atTerm->aSizeY) {
//          atTerm->aScroll += -(newShowY-atTerm->aShowY) * atTerm->aFontY;
			ScrollWindow(atTerm->aWnd, 0, -(newShowY-atTerm->aShowY)*atTerm->aFontY, NULL, NULL);
		} else {
			InvalidateRect(atTerm->aWnd, NULL, FALSE);
		}
	} else {
		/* scroll up */
		if ((atTerm->aShowY-newShowY) < atTerm->aSizeY) {
//          atTerm->aScroll += (atTerm->aShowY-newShowY) * atTerm->aFontY;
			ScrollWindow(atTerm->aWnd, 0, (atTerm->aShowY-newShowY)*atTerm->aFontY, NULL, NULL);
		} else {
			InvalidateRect(atTerm->aWnd, NULL, FALSE);
		}
	}
	atTerm->aShowY = newShowY;

	return;
}


int atMacroParse(unsigned char *pBuf, unsigned char *pMacro)
{
    int  i, j;
    unsigned char *p;
    int  setval;
    int  base, cnt;
    unsigned long rawChar;
    int  value;

    if ((!pBuf) || (!pMacro))  return 0;
    if (!pMacro[0])  return 0;

    p = pBuf;
    *p = 0x00;
    for (j = 0; (j < TBMACRO_MAXLEN) && (pMacro[j]); j++)  {
        if (pMacro[j] == '\\') {
            j++;
            if ((j < TBMACRO_MAXLEN) && (pMacro[j])) {
                setval = -1;
                if ((pMacro[j] == 'a') || (pMacro[j] == 'A')) {
                    setval = 0x07;
                } else if ((pMacro[j] == 'b') || (pMacro[j] == 'B')) {
                    setval = 0x08;
                } else if ((pMacro[j] == 't') || (pMacro[j] == 'T')) {
                    setval = 0x09;
                } else if ((pMacro[j] == 'n') || (pMacro[j] == 'N')) {
                    setval = 0x0A;
                } else if ((pMacro[j] == 'r') || (pMacro[j] == 'R')) {
                    setval = 0x0D;
                } else if (pMacro[j] == '\\') {
                    setval='\\';
                } else if ((pMacro[j] == '%') || (pMacro[j] == 'x') ||
                          ((pMacro[j] >= '0') && (pMacro[j] <= '9'))) {
                    /* interpret as a numeric value */
                    if ((j < TBMACRO_MAXLEN) && (pMacro[j])) {
                        if (pMacro[j] == '%') { /* decimal */
                            cnt = 3; /* # of letters to look for */
                            base = 10; /* our base */
                            j++;
                        } else if (pMacro[j] == 'x') { /* hexadecimal */
                            cnt = 2; /* # of letters to look for */
                            base = 16; /* our base */
                            j++;
                        } else if (pMacro[j] == 'b') { /* binary */
                            cnt = 8; /* # of letters to look for */
                            base = 2; /* our base */
                            j++;
                        } else { /* assume octal */
                            cnt = 3; /* # of letters to look for */
                            base = 8; /* our base */
                        }
                        rawChar = 0x00;
                        for (i = 0; i < cnt; i++) {
                            value = pMacro[j];
                            value &= 0x0FF;
                            if ((j >= TBMACRO_MAXLEN) || (!value))
                                break;
                            if ((value >= '0') && (value <= '9')) {
                                value -= '0';
                            } else {
                                if ((value >= 'a') && (value <= 'z')) {
                                    value -= 'a';
                                }
								else if ((value >= 'A') && (value <= 'Z')) {
                                    value -= 'A';
                                }
								else {
                                    break; /* invalid character - abort */
                                }
                                value += 10;
                            }
                            /* Check the validity of value */
                            if (value >= base)
                                break; /* invalid character */
                            rawChar *= base;
                            rawChar += value;
                            j++;
                        }
                        setval = (int)rawChar;
                        j--;
                    }
                } /* else ignore it */
                /* check if we have setval modified */
                if (setval != -1) {
                    *p = (char)setval;
                    p++;
                    *p = 0x00;
                }
            }
			else  {
                break; /* return, ignoring this partial (malformed) escape sequence */
            }
        }
		else {
            *p = pMacro[j];
            p++;
            *p = 0x00;
        }
    }
    j = p - pBuf;

    return j;
}


#define ATANSITERMPROCBUFSIZE 1024


/*
 * This proc allows a higher module to fetch data from the user input buffer,
 */
long vt100ReadInput(ATTERM *atTerm, unsigned char *pBuf, unsigned long pBufLen)
{
	unsigned long i;

	if (!atTerm)  return 0L;

	i = 0;
	while ((atTerm->aOutBufSockTail != atTerm->aOutBufHead) && (i < pBufLen)) {
		pBuf[i++] = *(atTerm->aOutBufSockTail);
		atTerm->aOutBufSockTail++;
		if ((unsigned long)(atTerm->aOutBufSockTail - atTerm->aOutBuf) >= atTerm->aOutBufSize)
			atTerm->aOutBufSockTail = atTerm->aOutBuf;
	}

	return i;
}


long atUnReadTermInput(ATTERM *atTerm, unsigned char *pBuf, unsigned long pBufLen)
{
	unsigned long i;

	if (!atTerm)  return 0L;

	i = 0;
	while ((atTerm->aOutBufSockTail != atTerm->aOutBufHead) && (i < pBufLen)) {
		atTerm->aOutBufSockTail--;
		if (atTerm->aOutBufSockTail < atTerm->aOutBuf)
			atTerm->aOutBufSockTail += atTerm->aOutBufSize;
		*(atTerm->aOutBufSockTail) = pBuf[i++];
	}

	return i;
}


/*
 * Inserts a string into the output buffer for the terminal.
 */
long atWriteTermInput(ATTERM *atTerm, unsigned char *pBuf, unsigned long pBufLen)
{
	unsigned long i;

	if (!atTerm)  return 0L;

	i = 0;
	while (i < pBufLen) {
		*(atTerm->aOutBufHead) = pBuf[i++];
		(atTerm->aOutBufHead)++;
		if ((unsigned long)(atTerm->aOutBufHead - atTerm->aOutBuf) >= atTerm->aOutBufSize)
			atTerm->aOutBufHead = atTerm->aOutBuf;
		if (atTerm->aOutBufHead == atTerm->aOutBufSockTail) {
			atTerm->aOutBufHead--;
			if (atTerm->aOutBufHead < atTerm->aOutBuf)
				atTerm->aOutBufHead = atTerm->aOutBuf + atTerm->aOutBufSize - 1;
			i--;
			break;
		}
	}

	return i;
}


/*
 * Processes a string as though the user typed it.
 */
void atSendTermInput(ATTERM *atTerm, unsigned char *pBuf, unsigned long pBufLen)
{
	unsigned long i;
	unsigned char outStr[256];

	if (!atTerm)  return;

	for (i = 0; i < pBufLen; i++) {
		/* do keyboard translation NOW */
		if (pBuf[i] == 0x7F)
			*outStr = atTerm->aDeleteChar;
		else if (pBuf[i] == 0x08)
			*outStr = atTerm->aBackSpaceChar;
		else
			*outStr = pBuf[i];

        /* insert into our output buffer */
        atWriteTermInput(atTerm, (unsigned char*)outStr, 1L);
        /* check echo to the screen */
        if (atTerm->aVLocalEcho) {
            if ((*outStr >= 0x20) && (*outStr<0x7F)) {
                vt100Send(atTerm, outStr, 1);
                (atTerm->aLE_BSCount)++;
			}
		    else {
                if ((*outStr == ASCII_CR) && (atTerm->aLFwithNLMode)) {
                    outStr[1] = ASCII_LF;
                    vt100Send(atTerm, (unsigned char *)outStr, 2);
                    atTerm->aLE_BSCount = 0;
				} else if (*outStr == ASCII_BS) {
                    if (atTerm->aLE_BSCount) {
                        vt100Send(atTerm, (unsigned char *)outStr, 1);
                        atTerm->aLE_BSCount--;
					} /* else don't echo */
				} else {
                    vt100Send(atTerm, (unsigned char *)outStr, 1);
                    atTerm->aLE_BSCount++;
				}
			}
		}
    }
    if (atTerm->aNotify)  (*(atTerm->aNotify))(atTerm, TBNOTIFY_USERINPUT);
}



void vt100SaveSettings(ATTERM *atTerm, int pSettings, char *pSection)
{
	char buf[20];
	char *sect;
	int  i;

	if (!atTerm)  return;

	if (pSection)
		sect = pSection;
	else
		sect = atTerm->aEnvSection;

	switch (pSettings) {
	  case TBNOTIFY_OPTIONS:
		enSaveLong(sect, "Emulation", atTerm->aEmulation);
		enSaveInt(sect, "Rows", atTerm->aSizeY);
		enSaveInt(sect, "Columns", atTerm->aSizeX);
		enSaveInt(sect, "LocalEcho", (unsigned int)atTerm->aLocalEcho);
		enSaveInt(sect, "Bell", (unsigned int)atTerm->aBell);
		enSaveInt(sect, "BackSpaceChar", (unsigned int)atTerm->aBackSpaceChar);
		enSaveInt(sect, "DeleteChar", (unsigned int)atTerm->aDeleteChar);
		enSaveInt(sect, "Scrollback", (unsigned long)(atTerm->aTermBuf->tBufY-atTerm->aTermBuf->tScrY-ATIB_TERMBUF_EXTRALINES));
		enSaveInt(sect, "AutoRightMargin", (unsigned int)atTerm->aTermBuf->tAutoRightMargin);
		break;

	  case TBNOTIFY_FONT:
		if (atTerm->aLogFontValid) {
			enSaveInt(sect, "LogFontValid", 1);
			enSaveLogFont(sect, &(atTerm->aLogFont));
		} else {
			enSaveInt(sect, "LogFontValid", 0);
		}
		break;

	  case TBNOTIFY_MACROS:
		for (i = 0; i < 48; i++) {
			sprintf(buf, "Macro%i", i+1);
			enSaveString(sect, buf, atTerm->aFnKeyMacro[i]);
		}
		break;

	  default:
		break;
	}
}


void vt100LoadSettings(ATTERM *atTerm, int pSettings, char *pSection)
{
	char    *sect;
	LOGFONT logFont;
	int     i;
	char    buf[20];
	unsigned long l_u;

	if (!atTerm)  return;

	if (pSection)
		sect = pSection;
	else
		sect = atTerm->aEnvSection;

	switch (pSettings) {
	  case TBNOTIFY_OPTIONS:
		atTerm->aEmulation = enRestoreLong(sect, "Emulation", atTerm->aEmulation);
		atTerm->aSizeY = enRestoreInt(sect, "Rows", 25);
		atTerm->aSizeX = enRestoreInt(sect, "Columns", 80);
		atTerm->aLocalEcho = (BOOL)enRestoreInt(sect, "LocalEcho", 1);
		atTerm->aBell = (BOOL)enRestoreInt(sect, "Bell", 1);
		atTerm->aBackSpaceChar = (unsigned char)enRestoreInt(sect, "BackSpaceChar", ASCII_BS);
		atTerm->aDeleteChar = (unsigned char)enRestoreInt(sect, "DeleteChar", ASCII_DEL);
		l_u = (unsigned int)enRestoreInt(sect, "Scrollback", 100);
		atAdjustTerm(atTerm, l_u);
		enSaveInt(sect, "Scrollback",
			(unsigned long)(atTerm->aTermBuf->tBufY - atTerm->aTermBuf->tScrY - ATIB_TERMBUF_EXTRALINES));
		atTerm->aTermBuf->tAutoRightMargin= (BOOL)enRestoreInt(sect, "AutoRightMargin", TRUE);
		break;

	  case TBNOTIFY_FONT:
		if (enRestoreInt(sect, "LogFontValid", 0)) {
			enRestoreLogFont(sect, &logFont);
			vt100ChangeFont(atTerm, NULL, &logFont);
		} else {
			vt100ChangeFont(atTerm, GetStockObject(SYSTEM_FIXED_FONT), NULL);
		}
		break;

	  case TBNOTIFY_MACROS:
		for (i = 0; i < 48; i++) {
			sprintf(buf, "Macro%i", i+1);
			enRestoreString(sect, buf, "", atTerm->aFnKeyMacro[i], TBMACRO_MAXLEN);
		}
		break;

	  default:
		break;
	}
}


void atResetTerm(ATTERM *atTerm)
{
	int i;

	atTerm->aVLocalEcho = FALSE;
	for (i = 0; i < 4; i++)
		atTerm->aLED[i] = FALSE;
	atTerm->aLFwithNLMode = TRUE;
	atTerm->aCursorKeyAppMode = FALSE;
	atTerm->aANSIMode       = TRUE;
	atTerm->aColumnMode     = TRUE;
	atTerm->aScrollingMode  = TRUE;
	atTerm->aScreenMode     = TRUE;
	atTerm->aWraparoundMode = TRUE;
	atTerm->aAutoRepeatMode = TRUE;
}



LRESULT CALLBACK  vt100WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static unsigned char buf[ATANSITERMPROCBUFSIZE];
	PAINTSTRUCT ps;           // paint structure
	RECT   rect;
	HDC    hDC;               // device context
	int    i, j, x, y, cnt;
	int    startX, endX;      // start & end lines
	int    startY, endY;      // start & end lines
	HFONT  oldFont;
	HPEN   oldPen;
	ATTERM *atTerm;
	unsigned long attrib;
	unsigned long rawChar, *rawCharP;
	unsigned char *p;
	HMENU  menu;
	POINT  point;


	atTerm = (ATTERM *)GetWindowLong(hWnd, ATWND_ATTERMSEG);

	switch (msg)  {
	  case WM_VSCROLL:
		DestroyCaret();
		switch (GET_WM_VSCROLL_CODE(wParam, lParam)) {
		  case SB_BOTTOM:
			atScrollTermTo(atTerm, 0, 0);
			break;

		  case SB_LINEDOWN:
			atScrollTermTo(atTerm, 1, 1);
			break;

		  case SB_LINEUP:
			atScrollTermTo(atTerm, 1, -1);
			break;

		  case SB_PAGEDOWN:
			atScrollTermTo(atTerm, 1, atTerm->aTermBuf->tScrY-1);
			break;

		  case SB_PAGEUP:
			atScrollTermTo(atTerm, 1, -(atTerm->aTermBuf->tScrY-1));
			break;

		  case SB_THUMBPOSITION:
		  case SB_THUMBTRACK:
			atScrollTermTo(atTerm, 0,(long)((short)(GET_WM_HSCROLL_POS(wParam, lParam))));
			UpdateWindow(hWnd);
			break;

		  case SB_TOP:
			atScrollTermTo(atTerm, 0, 0);
			break;

		  case SB_ENDSCROLL:
			SendMessage(hWnd, WM_SETFOCUS, 0, 0L);
			break;
		}
		SetScrollPos(hWnd, SB_VERT, atTerm->aShowY, TRUE);
		return 0L;

	  case WM_COMMAND:
		switch (GET_WM_COMMAND_ID(wParam, lParam)) {
		  case ID_EDIT_COPY:
			i = tbHilightSize(atTerm->aTermBuf);
			if (!i)
				break; /* nothing to copy */
			if (OpenClipboard(atTerm->aWnd)) {
				if (EmptyClipboard()) {
					HGLOBAL l_hGlobal;
					l_hGlobal = GlobalAlloc(GMEM_DDESHARE|GMEM_MOVEABLE,i+10); /* + 10 just to be sure we don't overrun */
					if (l_hGlobal) {
						p = (unsigned char*)GlobalLock(l_hGlobal);
						if (p) {
							tbCopyHilight(atTerm->aTermBuf, p);
							GlobalUnlock(l_hGlobal);
							SetClipboardData(CF_TEXT, (HANDLE)l_hGlobal);
						} else {
							GlobalFree(l_hGlobal);
						}
					}
				}
				CloseClipboard();
			}
			break;

		  case ID_EDIT_PASTE :
			if (IsClipboardFormatAvailable(CF_TEXT)) {
				if (OpenClipboard(atTerm->aWnd)) {
					HGLOBAL l_hGlobal;
					l_hGlobal = GetClipboardData(CF_TEXT);
					if (l_hGlobal) {
						unsigned char *l_pp;
						l_pp = (unsigned char*)GlobalLock(l_hGlobal);
						if (l_pp) {
							/* insert text */
							atSendTermInput(atTerm,l_pp,strlen((char*)l_pp));
							GlobalUnlock(l_hGlobal);
						}
					}
					CloseClipboard();
				}
			}
			break;

		  default:
			break;
		}
		break;

	  case WM_LBUTTONDBLCLK:
		break;

	  case WM_RBUTTONDOWN:
		/* Up our pop-up menu */
		menu = CreatePopupMenu();
		if (menu) {
			AppendMenu(menu, MF_ENABLED|MF_STRING, ID_EDIT_COPY, "&Copy\tCtrl+Ins");
			if (IsClipboardFormatAvailable(CF_TEXT))
				AppendMenu(menu,MF_ENABLED|MF_STRING, ID_EDIT_PASTE, "&Paste\tShift+Ins");
			else
				AppendMenu(menu,MF_DISABLED|MF_GRAYED,ID_EDIT_PASTE, "&Paste\tShift+Ins");
			if (GetCursorPos(&point)) {
				TrackPopupMenu(menu,TPM_TOPALIGN|TPM_LEFTALIGN|TPM_RIGHTBUTTON,
				point.x, point.y, 0, hWnd, NULL);
			}
			DestroyMenu(menu);
		}
		break;

	  case WM_LBUTTONDOWN:
		/* Start hilighting */
		if (!GetCursorPos(&point))
				break;
		ScreenToClient(atTerm->aWnd, &point);
		/* Erase any pre-existing hilight area */
		tbClearHilight(atTerm->aTermBuf);
		InvalidateRect(atTerm->aWnd, NULL, FALSE);
		/* Capture our mouse */
		SetCapture(hWnd);
		atTerm->aMouseHiliteCapture = TRUE;
		atTerm->aCaptureLastX = point.x;
		atTerm->aCaptureLastY = point.y;
		tbStartHilight(atTerm->aTermBuf, (point.x/atTerm->aFontX)+atTerm->aShowX,
				(point.y/atTerm->aFontY) + atTerm->aShowY);
		/* Generate a timer to check on scrolling */
		SetTimer(atTerm->aWnd, AT_TIMER_SCROLL, 100, NULL);
		break;

	  case WM_LBUTTONUP:
		if (atTerm->aMouseHiliteCapture) {
			/* Release our mouse */
			atTerm->aMouseHiliteCapture = FALSE;
			ReleaseCapture();
			KillTimer(atTerm->aWnd, 1);
		}
		break;

	  case WM_TIMER :
		switch (wParam) {
        case AT_TIMER_SCROLL:
          if (atTerm->aMouseHiliteCapture) {
            i = 0;
            if (atTerm->aCaptureLastY < 0) {
              i = atTerm->aCaptureLastY;
              if (i < -25) {
                i += 15;
                i /= 5;
              }
			  else {
                i = -1;
              }
              rect.left  = 0;
              rect.right = atTerm->aFontX * atTerm->aSizeX;
              rect.top   = 0;
              rect.bottom= ((-i)+1) * atTerm->aFontY;
            }
			else if (atTerm->aCaptureLastY > atTerm->aFontY * atTerm->aSizeY) {
              i = (atTerm->aCaptureLastY-atTerm->aFontY * atTerm->aSizeY);
              if (i > 25) {
                i -= 15;
                i /= 5;
              }
			  else {
                i = 1;
              }
              rect.left  = 0;
              rect.right = atTerm->aFontX * atTerm->aSizeX;
              rect.top   = atTerm->aFontY * ((atTerm->aSizeY-i-1));
              rect.bottom= atTerm->aFontY * atTerm->aSizeY;
            }
            if (i) {
              atScrollTermTo(atTerm, 1, i);
              j = ((atTerm->aCaptureLastY/atTerm->aFontY)+atTerm->aShowY);
              if (j < atTerm->aSizeY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES)
                j = atTerm->aSizeY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES;
              tbUpdateHilight(atTerm->aTermBuf,
                (atTerm->aCaptureLastX/atTerm->aFontX)+atTerm->aShowX, j);
              InvalidateRect(atTerm->aWnd, &rect, FALSE);
              SetScrollPos(hWnd,SB_VERT, atTerm->aShowY, TRUE);
              atUpdateCaret(atTerm);
            }
          }
		  else {
            KillTimer(atTerm->aWnd, wParam);
          }
          break;

        case AT_TIMER_FLASH:
          FlashWindow(atTerm->aWnd, FALSE);
          KillTimer(atTerm->aWnd, wParam);
          break;
        }
		break;

	  case WM_MOUSEMOVE:
		if (atTerm->aMouseHiliteCapture) {
			if (GetCursorPos(&point)) {
				/* Capture our mouse */
				ScreenToClient(atTerm->aWnd, &point);
				i = (point.x/atTerm->aFontX)+atTerm->aShowX;
				j = (point.y/atTerm->aFontY)+atTerm->aShowY;
				if (j < atTerm->aSizeY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES) {
					j = atTerm->aSizeY-atTerm->aTermBuf->tBufY+ATIB_TERMBUF_EXTRALINES;
					i = 0;
				}
				tbUpdateHilight(atTerm->aTermBuf, i, j);
				if (atTerm->aCaptureLastY > point.y) {
					rect.top = (point.y/atTerm->aFontY)*atTerm->aFontY;
					rect.bottom = (atTerm->aCaptureLastY/atTerm->aFontY+1)*atTerm->aFontY;
				}
				else {
					rect.top = (atTerm->aCaptureLastY/atTerm->aFontY)*atTerm->aFontY;
					rect.bottom = (point.y/atTerm->aFontY+1)*atTerm->aFontY;
				}
				rect.left = 0;
				rect.right = atTerm->aSizeX * atTerm->aFontX;
				InvalidateRect(hWnd, &rect, FALSE);
				atTerm->aCaptureLastX = point.x;
				atTerm->aCaptureLastY = point.y;
			}
		}
		break;

	  case WM_KEYDOWN:
		/* let's see if this is a function key macro */
		/* VIRTUAL KEY VALUES:
		 * ------------------
		 * PgUp   : 0x21
		 * PgDn   : 0x22
		 * End    : 0x23
		 * Home   : 0x24
		 * Ins    : 0x2D
		 * Del    : 0x2E
		 * F1-F12 : 0x70-0x7B
		 * CsrLft : 0x25
		 * CsrUp  : 0x26
		 * CsrRgt : 0x27
		 * CsrDn  : 0x28
		 * ------------------
		 */
		i = LOWORD(wParam);
		if ((i >= 0x70) && (i <= 0x7B)) {
			i -= 0x70;
			if (GetKeyState(VK_SHIFT) & 0x00008000)  i += 12;
			if (GetKeyState(VK_CONTROL)&0x00008000)  i += 24;
			j = atMacroParse(buf, atTerm->aFnKeyMacro[i]);
			if (j) {
				/* Key has a macro override assigned to it */
  				atSendTermInput(atTerm, buf, j);
				return 0L;
			}
			else {
				/* Send default function key sequence */
				if ((i >= 0) && (i <= 4)) {
					buf[0] = (unsigned char)ASCII_ESC;
					buf[1] = (unsigned char)'[';
					buf[2] = (unsigned char)'[';
					buf[3] = (unsigned char)('A'+i);
					atSendTermInput(atTerm, buf, 4);
				}
				else if ((i >= 5) && (i <= 11)) {
					if (i >= 10) i++;
					sprintf((char*)buf," [%02i~",i+12);
					buf[0] = (unsigned char)ASCII_ESC;
					atSendTermInput(atTerm, buf, 5);
				}
			}
		}
		else if (i == 0x21) { /* Page UP */
            if (GetKeyState(VK_SHIFT) & 0x00008000) {
				atScrollTermTo(atTerm, 1, -(atTerm->aTermBuf->tScrY-1));
				SetScrollPos(hWnd, SB_VERT, atTerm->aShowY, TRUE);
				atUpdateCaret(atTerm);
			}
			else {
				/* send PGUP */
				buf[0] = (unsigned char)ASCII_ESC;
				buf[1] = (unsigned char)'[';
				buf[2] = (unsigned char)'5';
				buf[3] = (unsigned char)'~';
				atSendTermInput(atTerm, buf, 4);
			}
		}
		else if (i == 0x22) { /* Page DOWN */
            if (GetKeyState(VK_SHIFT)&0x00008000) {
				atScrollTermTo(atTerm, 1, atTerm->aTermBuf->tScrY-1);
				SetScrollPos(hWnd, SB_VERT, atTerm->aShowY, TRUE);
				atUpdateCaret(atTerm);
			}
			else {
				/* send PGDN */
				buf[0] = (unsigned char)ASCII_ESC;
				buf[1] = (unsigned char)'[';
				buf[2] = (unsigned char)'6';
				buf[3] = (unsigned char)'~';
				atSendTermInput(atTerm, buf, 4);
			}
		}
		else if (i == 0x23) { /* END */
			/* send END */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'4';
			buf[3] = (unsigned char)'~';
			atSendTermInput(atTerm, buf,4);
		}
		else if (i == 0x24) { /* HOME */
			/* Send Home */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'H';
			atSendTermInput(atTerm, buf, 3);
//			/* Send HOME */
//			buf[0] = (unsigned char)ASCII_ESC;
//			buf[1] = (unsigned char)'[';
//			buf[2] = (unsigned char)'1';
//			buf[3] = (unsigned char)'~';
//			atSendTermInput(atTerm, buf, 4);
		}
		else if (i == 0x25) { /* Cursor LEFT */
			/* Send ASCII Crsr Left */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'D';
			atSendTermInput(atTerm,  buf, 3);
		}
		else if (i == 0x27) { /* Cursor RIGHT */
			/* Send ASCII Crsr Right */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'C';
			atSendTermInput(atTerm, buf,3);
		}
		else if (i == 0x26) { /* Cursor UP */
			/* Send ASCII Crsr Up */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'A';
			atSendTermInput(atTerm, buf,3);
		}
		else if (i == 0x28) { /* Cursor DOWN */
			/* Send ASCII Crsr Down */
			buf[0] = (unsigned char)ASCII_ESC;
			buf[1] = (unsigned char)'[';
			buf[2] = (unsigned char)'B';
			atSendTermInput(atTerm, buf, 3);
		}
		else if (i == 0x2D) { /* INS */
			if (GetKeyState(VK_SHIFT) & 0x00008000) {
				/* Paste */
				FORWARD_WM_COMMAND(hWnd, ID_EDIT_PASTE, NULL, 0, PostMessage);
			}
			else if (GetKeyState(VK_CONTROL) & 0x00008000) {
				/* Copy */
				FORWARD_WM_COMMAND(hWnd, ID_EDIT_COPY, NULL, 0, PostMessage);
			}
			else {
				/* Send INS */
				buf[0] = (unsigned char)ASCII_ESC;
				buf[1] = (unsigned char)'[';
				buf[2] = (unsigned char)'2';
				buf[3] = (unsigned char)'~';
				atSendTermInput(atTerm, buf, 4);
			}
		}
		else if (i == 0x2E) { /* DEL */
			/* pass on DEL keystroke */
			buf[0] = (unsigned char)ASCII_DEL;
			atSendTermInput(atTerm, buf, 1);
			return 0L;
		}
		break;

	  case WM_CHAR:
		buf[0] = (unsigned char)wParam;
		atSendTermInput(atTerm, buf, 1);
		atScrollTermTo(atTerm, 0, 0);
		return 0L;

	  case WM_PAINT:
		/* Check for scrolling */
		if (atTerm->aScroll) {
			ScrollWindow(atTerm->aWnd, 0, atTerm->aScroll, NULL, NULL);
			atTerm->aScroll = 0;
		}
		/* Paint window */
		if (GetUpdateRect(hWnd, &rect, TRUE)) { /* get new update region */
			hDC = BeginPaint(hWnd, &ps);
			oldFont = SelectObject(hDC, atTerm->aFont);
			/* Determine line numbers to update */
			startY = (rect.top/atTerm->aFontY) + atTerm->aShowY;
			endY   = ((rect.bottom-1)/atTerm->aFontY) + atTerm->aShowY;
			startX = (rect.left/atTerm->aFontX) + atTerm->aShowX;
			endX   = ((rect.right-1)/atTerm->aFontX) + atTerm->aShowX;
			if (startX > TB_GETXMAX(atTerm->aTermBuf))
				startX = TB_GETXMAX(atTerm->aTermBuf);
			if (endX > TB_GETXMAX(atTerm->aTermBuf))
				endX = TB_GETXMAX(atTerm->aTermBuf);
			if (startY > TB_GETYMAX(atTerm->aTermBuf))
				startY = TB_GETYMAX(atTerm->aTermBuf);
			if (endY > TB_GETYMAX(atTerm->aTermBuf))
				endY = TB_GETYMAX(atTerm->aTermBuf);
			/* Paint rest of screen, if required */
			rawCharP = TB_GETCHARADDR(atTerm->aTermBuf, startX, startY);
			if (!IsBadReadPtr(rawCharP, 1)) {
				rawChar = *rawCharP;
				attrib = rawChar & (TB_EFFECTMASK|TB_BCOLORMASK|TB_FCOLORMASK);
				attrib ^= 0xFFFFFFFFL;
				cnt = 0;
				for (y = startY; y <= endY; y++) {
					for (x = startX; x <= endX; x++) {
						/* Print this character */
						rawCharP = TB_GETCHARADDR(atTerm->aTermBuf, x, y);
						if (!IsBadReadPtr(rawCharP, 1)) {
							rawChar = *rawCharP;
						}
						else {
							rawChar = TB_MAKEEFFECT(0)|TB_MAKEBCOLOR(0)|TB_MAKEFCOLOR(7)|TB_MAKECHAR(0x20);
						}
						if (tbIsHilight(atTerm->aTermBuf, x, y)) {
							rawChar ^= (TB_BCOLORMASK|TB_FCOLORMASK);
						}
						if ((attrib != (rawChar & (TB_EFFECTMASK|TB_BCOLORMASK|TB_FCOLORMASK))) ||
							(cnt >= (ATANSITERMPROCBUFSIZE-1))) {
							if (cnt) {
								if (atTerm->aFontUnderline) {
									/* check for underline */
									if (attrib & TB_MAKEEFFECT(AT_FONTEFFECT_UNDERLINE)) {
										SelectObject(hDC, atTerm->aFontUnderline);
										TextOut(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
													   (y-atTerm->aShowY) * atTerm->aFontY,
													   (LPCSTR)buf, cnt);
										SelectObject(hDC, atTerm->aFont);
									}
									else {
										TextOut(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
													   (y-atTerm->aShowY) * atTerm->aFontY,
													   (LPCSTR)buf,cnt);
									}
								}
								else {
									TextOut(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
												   (y-atTerm->aShowY) * atTerm->aFontY,
												   (LPCSTR)buf,cnt);
									/* Check for underline */
									if (attrib & TB_MAKEEFFECT(AT_FONTEFFECT_UNDERLINE)) {
										oldPen = SelectObject(hDC,
												atTerm->aTermBuf->tColorPen[i]);
										MoveToEx(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
														(y-atTerm->aShowY+1) * atTerm->aFontY-1, NULL);
										LineTo(hDC, (x-atTerm->aShowX+1) * atTerm->aFontX,
													  (y-atTerm->aShowY+1)*atTerm->aFontY-1);
										oldPen = SelectObject(hDC, oldPen);
									}
								}
								cnt = 0;
							}

							i = TB_EXTFCOLOR(rawChar);
							j = TB_EXTBCOLOR(rawChar);
							if (TB_EXTEFFECT(rawChar) & AT_FONTEFFECT_BOLD) {
								i |= 0x08;
							}
							if (TB_EXTEFFECT(rawChar) & AT_FONTEFFECT_BLINK) {
								j |= 0x08;
							}
							if (TB_EXTEFFECT(rawChar) & AT_FONTEFFECT_REVERSE) {
								if (i == 0x08)
									i = 0x00;
								else
									i ^= 0x07;
								if (j == 0x08)
									j = 0x00;
								else
									j ^= 0x07;
							}
							SetTextColor(hDC, atTerm->aTermBuf->tColorBank[i]);
							SetBkColor(hDC, atTerm->aTermBuf->tColorBank[j]);
							attrib = (rawChar & (TB_EFFECTMASK|TB_BCOLORMASK|TB_FCOLORMASK));
						}
						buf[cnt] = (char)TB_EXTCHAR(rawChar);
						cnt++;
					}
					if (cnt) {
						if (atTerm->aFontUnderline) {
							/* Check for underline */
							if (attrib & TB_MAKEEFFECT(AT_FONTEFFECT_UNDERLINE)) {
								SelectObject(hDC, atTerm->aFontUnderline);
								TextOut(hDC,(x-cnt-atTerm->aShowX) * atTerm->aFontX,
											  (y-atTerm->aShowY) * atTerm->aFontY,
											  (LPCSTR)buf, cnt);
								SelectObject(hDC, atTerm->aFont);
							}
							else {
								TextOut(hDC, (x-cnt-atTerm->aShowX)*atTerm->aFontX,
											   (y-atTerm->aShowY)*atTerm->aFontY,
											   (LPCSTR)buf, cnt);
							}
						}
						else {
							TextOut(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
										   (y-atTerm->aShowY) * atTerm->aFontY,
										   (LPCSTR)buf, cnt);
							/* Check for underline */
							if (attrib & TB_MAKEEFFECT(AT_FONTEFFECT_UNDERLINE)) {
								oldPen = SelectObject(hDC, atTerm->aTermBuf->tColorPen[i]);
								MoveToEx(hDC, (x-cnt-atTerm->aShowX) * atTerm->aFontX,
												(y-atTerm->aShowY+1) * atTerm->aFontY-1,NULL);
								LineTo(hDC, (x-atTerm->aShowX+1) * atTerm->aFontX,
											  (y-atTerm->aShowY+1)*atTerm->aFontY-1);
								oldPen = SelectObject(hDC, oldPen);
							}
						}
						cnt = 0;
					}
				}
			}
			EndPaint(hWnd, &ps);
			SelectObject(hDC, oldFont);
		}
		atUpdateCaret(atTerm);
		return 0L;

	  case WM_SETFOCUS:
		CreateCaret(hWnd, NULL, atTerm->aFontX, atTerm->aFontY);
		atUpdateCaret(atTerm);
		return 0L;

	  case WM_KILLFOCUS:
		DestroyCaret();
		return 0L;

	  case WM_DESTROY:
		enSaveInt(atTerm->aEnvSection, "Rows", atTerm->aSizeY);
		enSaveInt(atTerm->aEnvSection, "Columns", atTerm->aSizeX);
		enSaveWindowPlacement(atTerm->aEnvSection, atTerm->aWnd);
		if (atTerm->aNotify)  (*(atTerm->aNotify))(atTerm, ATNOTIFY_DESTROY);
		return 0L;

#ifdef BLUM_BLUM
	  case WM_WINDOWPOSCHANGING:
		/*
		 * Adjust our cx and cy so that our window is perfectly
		 * aligned with our row/column borders and we don't have any
		 * columns/rows half-displayed.
		 */
		if (((WINDOWPOS*)(lParam))->flags&SWP_NOSIZE)
			break;
		if (IsIconic(hWnd))
			break;

		i = (((WINDOWPOS*)(lParam))->cx) - 2 * GetSystemMetrics(SM_CXFRAME) - GetSystemMetrics(SM_CXVSCROLL) + GetSystemMetrics(SM_CXBORDER);
		j = (((WINDOWPOS*)(lParam))->cy) - 2 * GetSystemMetrics(SM_CYFRAME) - GetSystemMetrics(SM_CYCAPTION);

		/* Calculate our closest dimentions */
		x = (i + (atTerm->aFontX/2)) / atTerm->aFontX;
		y = (j + (atTerm->aFontY/2)) / atTerm->aFontY;
		if (x < AT_MIN_X)  x = AT_MIN_X;
		if (y < AT_MIN_Y)  y = AT_MIN_Y;

		/* Get column and row "remainders */
		i -= (x * atTerm->aFontX);
		j -= (y * atTerm->aFontY);

		/* Make changes to the WINDOWPOS struct */
		((WINDOWPOS*)(lParam))->cx -= i;
		((WINDOWPOS*)(lParam))->cy -= j;

		/* and then let the window handler continue it's job */
		break;
#endif
	  case WM_SIZE:
		if (IsIconic(hWnd))
			break;
        atAdjustTerm(atTerm, -1);
        atUpdateCaret(atTerm);
//        atTerm->aSizeX = LOWORD(lParam) / atTerm->aFontX;
//        atTerm->aSizeY = HIWORD(lParam) / atTerm->aFontY;
		break;

	  default:
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}



BOOL vt100InitClass(HINSTANCE hInst)
{
	WNDCLASS  wc;

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = vt100WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 4;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_VT100));
	wc.hCursor = LoadCursor(NULL, IDC_IBEAM);
	wc.hbrBackground = GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = AT_WIND_CLASS_NAME;

	return RegisterClass(&wc);
}


ATTERM *vt100CreateWnd(HINSTANCE inst, HWND parent,long bufX, long bufY, long scrX, long scrY, char *envSect)
{
	ATTERM  *term;
	HGLOBAL memory;
	long    i;
	int     x1, x2, y1, y2, xExtra, yExtra;
	WINDOWPLACEMENT *wp;

	/* Allocate ATTERM structure */
	memory = GlobalAlloc(GHND|GMEM_NOCOMPACT, sizeof(ATTERM));
	if (!memory) return NULL;

	term = (ATTERM*)GlobalLock(memory);
	if (!term) {
        GlobalFree(memory);
        return NULL;
    }
	/* Setup values and defaults */
	term->aMemory    = memory;
	term->aInst      = inst;
	term->aEmulation = AT_EMU_VT100;
    term->aSizeX     = enRestoreInt(envSect, "Columns", scrX);
    term->aSizeY     = enRestoreInt(envSect, "Rows", scrY);
	term->aShowX     = term->aShowY = 0;
	term->aState     = atGroundTable;
	strncpy(term->aEnvSection, envSect, TBSTR_ENVSECTION);
	term->aEnvSection[TBSTR_ENVSECTION-1] = '\0';
	term->aFont      = NULL;
	term->aFontUnderline = NULL;
	term->aLogFontValid  = FALSE;
	strncpy(term->aTitle, AT_DEFAULT_TITLE, TBSTR_TITLE_LEN);
	term->aTitle[TBSTR_TITLE_LEN-1] = '\0';
	term->aLocalEcho     = FALSE;
	term->aLFwithNLMode  = TRUE;
	term->aBackSpaceChar = ASCII_BS;
	term->aDeleteChar    = ASCII_DEL;
	term->aNotify        = NULL;
	term->aLE_BSCount    = 0L;
	term->aBell          = TRUE;
	term->aScroll        = 0L;

	/* ECMA-48 Codes and default Modes */
	term->aSIMD = 0;
	term->aSLH  = 0;

	/* Clear function key macros */
	for (i = 0; i < 12; i++)
		term->aFnKeyMacro[i][0] = 0;

	/* Allocat termbuf */
	term->aTermBuf = tbAllocTermBuf((unsigned long)(term), bufX, bufY+ATIB_TERMBUF_EXTRALINES, scrX, scrY);
    if (!term->aTermBuf) {
        GlobalUnlock(memory);
        GlobalFree(memory);
        return NULL;
    }

    /* Setup termbuf */
    ((void*)(term->aTermBuf->tNotifyFn)) = ((void*)(atNotifyFn));

    /* Allocate output buffer */
    term->aOutBufMemory = GlobalAlloc(GHND|GMEM_NOCOMPACT, 8192);
    if (!term->aOutBufMemory) {
        tbFreeTermBuf(term->aTermBuf);
        GlobalUnlock(memory);
        GlobalFree(memory);
        return NULL;
    }
    term->aOutBuf = (unsigned char*)GlobalLock(term->aOutBufMemory);
    if (!term->aOutBuf) {
        GlobalFree(term->aOutBufMemory);
        tbFreeTermBuf(term->aTermBuf);
        GlobalUnlock(memory);
        GlobalFree(memory);
        return NULL;
    }
	term->aOutBufSize = (unsigned long)GlobalSize(term->aOutBufMemory);
	term->aOutBufSockTail = term->aOutBufHead = term->aOutBuf;

	/* Create terminal window */
	term->aWnd = CreateWindow(AT_WIND_CLASS_NAME, AT_DEFAULT_TITLE,
     			 WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_VSCROLL,
				 CW_USEDEFAULT, CW_USEDEFAULT,
				 CW_USEDEFAULT, CW_USEDEFAULT,
				 parent, NULL, inst, NULL);
    if (!term->aWnd) {
        GlobalUnlock(term->aOutBufMemory);
        GlobalFree(term->aOutBufMemory);
        tbFreeTermBuf(term->aTermBuf);
        GlobalUnlock(memory);
        GlobalFree(memory);
        return NULL;
    }

    /* Setup extra window memory */
    SetWindowLong(term->aWnd, ATWND_ATTERMSEG, (LONG)term);

	/* Setup font */
	if (enRestoreInt(envSect, "LogFontValid", 0)) {
		LOGFONT logFont;

		enRestoreLogFont(envSect, &logFont);
		atSetFont(term, NULL, &logFont);
	} else {
		atSetFont(term, GetStockObject(SYSTEM_FIXED_FONT), NULL);
	}

    /* Calculate default dimentions -- default is AT_DEFAULT_X x AT_DEFAULT_Y */
    /* xExtra=width of window edges, scrollbar (assumed if defaults used), etc */
    xExtra = 2 * GetSystemMetrics(SM_CXFRAME) + GetSystemMetrics(SM_CXVSCROLL) + GetSystemMetrics(SM_CXBORDER);
    yExtra = 2 * GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION);
    if ((AT_DEFAULT_X * term->aFontX+xExtra) > GetSystemMetrics(SM_CXSCREEN)) {
        x1 = GetSystemMetrics(SM_CXSCREEN)*1/20;
        x2 = GetSystemMetrics(SM_CXSCREEN)*19/20;
    } else {
        /* the "/3" at the end is to up it in the upper-left corner, slightly */
        x1 = (GetSystemMetrics(SM_CXSCREEN) - (AT_DEFAULT_X*term->aFontX) - xExtra)/3;
        x2 = x1 + AT_DEFAULT_X * term->aFontX + xExtra;
    }
    if ((AT_DEFAULT_Y * term->aFontY+yExtra) > GetSystemMetrics(SM_CXSCREEN)) {
        y1 = (GetSystemMetrics(SM_CYSCREEN)*1/10);
        y2 = GetSystemMetrics(SM_CYSCREEN)*9/10;
    } else {
        /* the "/3" at the end is to up it in the upper-left corner, slightly */
        y1 = (GetSystemMetrics(SM_CYSCREEN) - (AT_DEFAULT_Y*term->aFontY) - yExtra)/3;
        y2 = y1 + AT_DEFAULT_Y * term->aFontY + yExtra;
    }
    wp = enRestoreWindowPlacement(term->aEnvSection, x1, y1, x2, y2);
    SetWindowPlacement(term->aWnd, wp);
    atAdjustTerm(term, -1);
    tbClearScreen(term->aTermBuf);
    term->aSLL = TB_GETYMAX(term->aTermBuf);

    atResetTerm(term);

    return term;
}

